#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Module to concatenate LCIO files.

:author: Ching Bon Lam
:since: Dec 17, 2011
"""

from __future__ import absolute_import
import os

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities.Subprocess import shellCall

from ILCDIRAC.Core.Utilities.PrepareLibs import removeLibc
from ILCDIRAC.Core.Utilities.resolvePathsAndNames import getProdFilename, resolveIFpaths
from ILCDIRAC.Workflow.Modules.ModuleBase import ModuleBase

__RCSID__ = '$Id$'
LOG = gLogger.getSubLogger(__name__)


class LCIOSplit(ModuleBase):
  """LCIO split module."""

  def __init__(self):

    super(LCIOSplit, self).__init__()

    self.STEP_NUMBER = ''
    self.nbEventsPerSlice = 0
    # Step parameters
    self.prod_outputdata = []
    self.applicationName = "lcio"
    #
    self.listoutput = {}
    LOG.info("%s initialized" % (self.__str__()))

  def applicationSpecificInputs(self):
    """Resolve LCIO concatenate specific parameters, called from ModuleBase."""

    if not self.OutputFile:
      return S_ERROR('No output file defined')

    if 'IS_PROD' in self.workflow_commons:
      if self.workflow_commons["IS_PROD"]:
        if 'ProductionOutputData' in self.workflow_commons:
          self.prod_outputdata = self.workflow_commons['ProductionOutputData'].split(";")
          for obj in self.prod_outputdata:
            if obj.lower().count("_sim_") or obj.lower().count("_rec_") or obj.lower().count("_dst_"):
              self.OutputFile = os.path.basename(obj)
        else:
          self.OutputFile = getProdFilename(self.OutputFile, int(self.workflow_commons["PRODUCTION_ID"]),
                                            int(self.workflow_commons["JOB_ID"]))

    if not len(self.InputFile) and len(self.InputData):
      for files in self.InputData:
        if files.lower().find(".slcio") > -1:
          self.InputFile.append(files)

    if 'listoutput' in self.step_commons:
      if len(self.step_commons['listoutput']):
        self.listoutput = self.step_commons['listoutput'][0]

    return S_OK('Parameters resolved')

  def execute(self):
    """Execute the module, called by JobAgent."""
    # Checks
    resultRIV = self.resolveInputVariables()
    if not resultRIV['OK']:
      LOG.error("Failed to resolve input variables", resultRIV['Message'])
      return resultRIV

    if not self.platform:
      return S_ERROR('No ILC platform selected')

    if "LCIO" not in os.environ:
      LOG.error("Environment variable LCIO was not defined, cannot do anything")
      return S_ERROR("Environment variable LCIO was not defined, cannot do anything")

    if len(self.InputFile):
      res = resolveIFpaths(self.InputFile)
      if not res['OK']:
        LOG.error("Missing slcio file!")
        self.setApplicationStatus('LCIOSplit: missing input slcio file')
        return S_ERROR('Missing slcio file!')
      runonslcio = res['Value'][0]
    else:
      return S_OK("No files found to process")

    removeLibc(os.path.join(os.environ["LCIO"], "lib"))

    # Setting up script

    LD_LIBRARY_PATH = os.path.join("$LCIO", "lib")
    if 'LD_LIBRARY_PATH' in os.environ:
      LD_LIBRARY_PATH += ":" + os.environ['LD_LIBRARY_PATH']

    PATH = "$LCIO/bin"
    if 'PATH' in os.environ:
      PATH += ":" + os.environ['PATH']

    scriptContent = """
#!/bin/sh

################################################################################
# Dynamically generated script by LCIOConcatenate module                       #
################################################################################

declare -x LD_LIBRARY_PATH=%s
declare -x PATH=%s

lcio split -i %s -n %s

exit $?

""" % (
        LD_LIBRARY_PATH,
        PATH,
        runonslcio,
        self.nbEventsPerSlice
        )

    # Write script to file

    scriptPath = 'LCIOSplit_%s_Run_%s.tcl' % (self.applicationVersion, self.STEP_NUMBER)

    if os.path.exists(scriptPath):
      os.remove(scriptPath)

    script = open(scriptPath, 'w')
    script.write(scriptContent)
    script.close()

    # Setup log file for application stdout

    if os.path.exists(self.applicationLog):
      os.remove(self.applicationLog)

    # Run code

    os.chmod(scriptPath, 0o755)

    command = '"./%s"' % (scriptPath)

    self.setApplicationStatus('LCIOSplit %s step %s' % (self.applicationVersion, self.STEP_NUMBER))
    self.stdError = ''

    result = shellCall(0,
                        command,
                        callbackFunction=self.redirectLogOutput,
                        bufferLimit=20971520
                      )

    # Check results
    status = result['Value'][0]

    if not os.path.exists(self.applicationLog):
      LOG.error("Cannot access log file, cannot proceed")
      return S_ERROR("Failed reading the log file")

    baseinputfilename = os.path.basename(runonslcio).split(".slcio")[0]
    output_file_base_name = ''
    if self.OutputFile:
      output_file_base_name = self.OutputFile.split('.slcio')[0]
    LOG.info("Will rename all files using '%s' as base." % output_file_base_name)
    numberofeventsdict = {}
    fname = ''
    with open(self.applicationLog, "r") as logf:
      for line in logf:
        line = line.rstrip()
        if line.count(baseinputfilename):
          # First, we need to rename those guys
          current_file = os.path.basename(line).replace(".slcio", "")
          current_file_extension = current_file.replace(baseinputfilename, "")
          newfile = output_file_base_name + current_file_extension + ".slcio"
          os.rename(line, newfile)
          fname = newfile
          numberofeventsdict[fname] = 0
        elif line.count("events"):
          numberofeventsdict[fname] = int(line.split()[0])

    LOG.verbose("Number of eventsdict dict: %s" % numberofeventsdict)

    # Now update the workflow_commons dict with the relation between filename and number of events: needed for
    # the registerOutputData
    self.workflow_commons['file_number_of_event_relation'] = numberofeventsdict
    if self.listoutput:
      outputlist = []
      for fileName in numberofeventsdict:
        item = {}
        item['outputFile'] = fileName
        item['outputPath'] = self.listoutput['outputPath']
        item['outputDataSE'] = self.listoutput['outputDataSE']
        outputlist.append(item)
      self.step_commons['listoutput'] = outputlist

    # Not only the step_commons must be updated
    if 'ProductionOutputData' in self.workflow_commons:
      proddata = self.workflow_commons['ProductionOutputData'].split(";")
      finalproddata = []
      this_split_data = ''
      for item in proddata:
        if not item.count(output_file_base_name):
          finalproddata.append(item)
        else:
          this_split_data = item
      path = os.path.dirname(this_split_data)
      for fileName in numberofeventsdict:
        finalproddata.append(os.path.join(path, fileName))
      self.workflow_commons['ProductionOutputData'] = ";".join(finalproddata)

    LOG.info("Status after the application execution is %s" % str(status))
    self.listDir()
    return self.finalStatusReport(status)



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Checks the validity of the provided XML files at job submission.

Created on Feb 24, 2011

:author: S. Poss
:since: Feb 24, 2011
"""

from __future__ import absolute_import
__RCSID__ = "$Id$"

from DIRAC import S_OK, S_ERROR
from xml.etree.ElementTree import ElementTree


def checkXMLValidity(xmlfile):
  """Check that the xml parsing of the specified xml will not fail when running on the GRID.

  :param str xmlfile: path to xml file
  :returns: :func:`S_OK() <DIRAC:DIRAC.Core.Utilities.ReturnValues.S_OK>`, :func:`~DIRAC:DIRAC.Core.Utilities.ReturnValues.S_ERROR`
  """
  tree = ElementTree()
  try:
    tree.parse(xmlfile)
  except Exception as x:
    return S_ERROR("Found problem in file %s: %s %s" % (xmlfile, Exception, x))

  return S_OK()



#!/usr/env python
#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

"""Make a RemovalRequest for the new System."""
from __future__ import print_function
from __future__ import absolute_import
__RCSID__ = "$Id$"

from DIRAC.Core.Base import Script
Script.parseCommandLine()

from DIRAC.RequestManagementSystem.Client.Request import Request

from DIRAC.RequestManagementSystem.Client.File import File
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.private.RequestValidator import RequestValidator
from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient


def myRequest():
  """Create a request and put it to the db."""

  request = Request()
  request.RequestName = 'myAwesomeRemovalRequest.xml'
  request.JobID = 0
  request.SourceComponent = "myScript"

  remove = Operation()
  remove.Type = "RemoveFile"

  lfn = "/ilc/user/s/sailer/test.txt"
  rmFile = File()
  rmFile.LFN = lfn
  remove.addFile(rmFile)

  request.addOperation(remove)
  isValid = RequestValidator().validate(request)
  if not isValid['OK']:
    raise RuntimeError("Failover request is not valid: %s" % isValid['Message'])
  else:
    print("It is a GOGOGO")
    requestClient = ReqClient()
    result = requestClient.putRequest(request)
    print(result)


if __name__ == "__main__":
  myRequest()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Contains the list of models and their properties.

:author: S. Poss
:since: Jul 07, 2011
"""
from __future__ import absolute_import
__RCSID__ = "$Id$"

from DIRAC import S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations


class GeneratorModels(object):
  """Contains the list of known models."""

  def __init__(self):
    self.ops = Operations()
    self.models = {}
    res = self.ops.getOptionsDict("/Models")
    if res['OK']:
      self.models = res['Value']

  def hasModel(self, model):
    """Check that specified model exists."""
    if model in self.models:
      return S_OK()
    else:
      return S_ERROR("Model %s is not defined, use any of %s" % (model, list(self.models.keys())))

  def getFile(self, model):
    """Return the proper model file (usually LesHouches)"""
    res = self.hasModel(model)
    if not res['OK']:
      return res
    if not self.models[model]:
      return S_ERROR("No file attached to model %s" % model)
    return S_OK(self.models[model])

  def getParamsForWhizard(self, model):
    """When creating the final file, this is needed to get the parameters for the SM."""
    params = ''
    if model == 'sm':
      params = """<GF type="float" value="1.16639E-5">
<!-- Fermi constant -->
</GF>
<mZ type="float" value="91.1882">
<!-- Z-boson mass -->
</mZ>
<mW type="float" value="80.419">
<!-- W-boson mass -->
</mW>
<mH type="float" value="120">
<!-- Higgs mass -->
</mH>
<alphas type="float" value="0.1178">
<!-- Strong coupling constant alpha_s(MZ) -->
</alphas>
<me type="float" value="0.">
<!-- electron mass -->
</me>
<mmu type="float" value="0.1066">
<!-- muon mass -->
</mmu>
<mtau type="float" value="1.777">
<!-- tau-lepton mass -->
</mtau>
<ms type="float" value="0.">
<!-- s-quark mass -->
</ms>
<mc type="float" value="0.54">
<!-- c-quark mass -->
</mc>
<mb type="float" value="2.9">
<!-- b-quark mass -->
</mb>
<mtop type="float" value="174">
<!-- t-quark mass -->
</mtop>
<wtop type="float" value="1.523">
<!-- t-quark width -->
</wtop>
<wZ type="float" value="2.443">
<!-- Z-boson width -->
</wZ>
<wW type="float" value="2.049">
<!-- W-boson width -->
</wW>
<wH type="float" value="0.3605E-02">
<!-- Higgs width -->
</wH>
<vckm11 type="float" value="0.97383">
<!-- Vud -->
</vckm11>
<vckm12 type="float" value="0.2272">
<!-- Vus -->
</vckm12>
<vckm13 type="float" value="0.00396">
<!-- Vub -->
</vckm13>
<vckm21 type="float" value="-0.2271">
<!-- Vcd -->
</vckm21>
<vckm22 type="float" value="0.97296">
<!-- Vcs -->
</vckm22>
<vckm23 type="float" value="0.04221">
<!-- Vcb -->
</vckm23>
<vckm31 type="float" value="0.00814">
<!-- Vtd -->
</vckm31>
<vckm32 type="float" value="-0.04161">
<!-- Vts -->
</vckm32>
<vckm33 type="float" value="0.99910">
<!-- Vtb -->
</vckm33>
<khgaz type="float" value="1.000">
<!-- anomaly Higgs coupling K factors -->
</khgaz>
<khgaga type="float" value="1.000">
<!-- anomaly Higgs coupling K factors -->
</khgaga>
<khgg type="float" value="1.000">
<!-- anomaly Higgs coupling K factors -->
</khgg>
"""
    else:
      params = """
      """
    return S_OK(params)



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Test for the SoftwareVersions Executor."""

from __future__ import absolute_import
import pytest
from mock import MagicMock, patch

from DIRAC import gLogger, S_OK


@pytest.fixture
def svExecutor(mocker):
  """Mock the SoftwareVersions executor."""
  class FakeExecutorModule(object):
    log = gLogger

    def __init__(self, *args, **kwargs):
      self.log = gLogger
      self.ex_optimizerName = lambda: "optimus prime"

    @classmethod
    def ex_getOption(cls, *args, **kwargs):
      return {'BanLists': ['1', '2'],
              '1Reason': 'BadSoftware',
              '1Sites': ['Site1', 'Site2'],
              '2Reason': 'BadVersion',
              '2Sites': ['Site2', 'Site3'],
              }[args[0]]

    def setNextOptimizer(self, *args, **kwargs):
      return S_OK()

  mocker.patch('ILCDIRAC.WorkloadManagementSystem.Executor.SoftwareVersions.OptimizerExecutor',
               new=MagicMock(spec="DIRAC.WorkloadManagementSystem.Executor.Base.OptimizerExecutor.OptimizerExecutor"))

  from ILCDIRAC.WorkloadManagementSystem.Executor import SoftwareVersions
  executorClass = SoftwareVersions.SoftwareVersions
  patchBase = patch.object(executorClass, '__bases__', (FakeExecutorModule,))
  with patchBase:
    patchBase.is_local = True
    SoftwareVersions.SoftwareVersions.initializeOptimizer()
    theExecutor = SoftwareVersions.SoftwareVersions()
    theExecutor._updateBanLists()
    theExecutor._updateBanLists = MagicMock()

  theExecutor.setNextOptimizer = MagicMock()
  return theExecutor


@pytest.fixture
def aJobState():
  """Return a jobState mock."""
  js = MagicMock(name="JobState")
  jm = MagicMock(name="JobManifest")

  def _jsOptions(*args, **kwargs):
    return {'SoftwarePackages': ['BadSoftware', 'BadVersion'],
            'BannedSites': [],
            'BannedSite': [],
            }[args[0]]
  jm.getOption = _jsOptions
  jm.setOption = MagicMock(name="setOption")

  js.getManifest.return_value = S_OK(jm)
  js.JM = jm  # for fast access
  return js


def test_optimizeJob(svExecutor, aJobState):
  """Test the optimizeJob function."""
  gLogger.setLevel('DEBUG')
  assert 'BadSoftware' in svExecutor._SoftwareVersions__softToBanned
  res = svExecutor.optimizeJob(1234, aJobState)
  assert res['OK']
  aJobState.JM.setOption.assert_called_with('BannedSites', 'Site1, Site2, Site3')



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Obtain the related steering file directory given a certain software name, version and platform.

Created on Feb 10, 2012

:author: Stephane Poss
:since: Feb 10, 2012
"""

from __future__ import absolute_import
__RCSID__ = "$Id$"

from DIRAC import S_OK, S_ERROR
from ILCDIRAC.Core.Utilities.CombinedSoftwareInstallation import getSoftwareFolder, checkCVMFS
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from ILCDIRAC.Core.Utilities.TARsoft import check


def getSteeringFileDirName(platform, application, applicationVersion):
  """Locate the path of the steering file directory assigned to the specified application."""
  ops = Operations()
  version = ops.getValue('/AvailableTarBalls/%s/%s/%s/Dependencies/steeringfiles/version' % (platform,
                                                                                             application,
                                                                                             applicationVersion), '')
  if not version:
    return S_ERROR("Could not find attached SteeringFile version")

  return getSteeringFileDir(platform, version)


def getSteeringFileDir(platform, version):
  """Return directly the directory, without passing by the dependency resolution."""
  res = checkCVMFS(platform, ['steeringfiles', version])
  if res['OK']:
    return S_OK(res['Value'][0])
  # Here means CVMFS is not defined, so we need to rely on the tar ball
  res = getSoftwareFolder(platform, 'steeringfiles', version)
  if not res['OK']:
    return res
  mySoftDir = res['Value']
  # check that all the files are there: software is not corrupted.
  res = check('steeringfiles.%s' % version, '.', [mySoftDir])
  if not res['OK']:
    return res
  return S_OK(mySoftDir)



#!/usr/bin/env python
#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Tests the ReportErrors WorkflowModule."""

from __future__ import absolute_import
import unittest
from collections import defaultdict
from mock import MagicMock as Mock, call, patch

from ILCDIRAC.Workflow.Modules.ReportErrors import ReportErrors
from Tests.Utilities.GeneralUtils import MatchStringWith

__RCSID__ = "$Id$"
MODULE_NAME = 'ILCDIRAC.Workflow.Modules.ReportErrors'


class TestReportErrors(unittest.TestCase):
  """Test ReportErrors."""

  def setUp(self):
    """Set up the tests."""
    self.log = Mock()
    self.log.info = Mock(name="LogInfo")
    self.log.error = Mock(name="LogError")

    self.patches = [patch('%s.LOG' % MODULE_NAME, new=self.log)]

    for patcher in self.patches:
      patcher.start()

    self.repErr = ReportErrors()
    self.repErr.workflow_commons = {}

  def tearDown(self):
    """Clean up test resources."""
    for patcher in self.patches:
      patcher.stop()

  def test_execute(self):
    """Test the execute function."""
    res = self.repErr.execute()
    self.assertTrue(res['OK'])
    self.log.info.assert_called_with(MatchStringWith('No errors encountered'))

    errorKey = "%s_%s" % ('appname', 'appver')
    message = 'something really bad'
    stdError = 'Segmentation Violation'
    self.repErr.workflow_commons.setdefault('ErrorDict', defaultdict(list))[errorKey].extend([message, stdError])

    res = self.repErr.execute()

    self.assertTrue(res['OK'])
    calls = [call(errorKey, 'something really bad'),
             call(errorKey, 'Segmentation Violation')]
    self.log.error.assert_has_calls(calls, any_order=True)



"""This is the the script for chaining many pandora photon likelihood files into a single likelihood file.

Input: ['InputFile1.xml', 'InputFile2.xml']
Ouput: OutputFile.xml

Tested with Pandora v02-08-02. If photon reconstruction algorithm in LCContent is not changed, then this script should
be valid
"""


# !/usr/bin/python

from __future__ import absolute_import
import xml.etree.ElementTree as ET
from DIRAC import S_OK, S_ERROR
from six.moves import zip


def getFixedXMLTree(inputFile):
  """Get fixed xml files by adding a top element.

  :param str inputFile: a string of input file
  :returns: an elementtree with fixed top element
  """
  content = ""
  with open(inputFile, "r") as f:
    content = f.read()
  # fix xml; Added top element in order to use the xml tree structure
  content = "<HEAD>\n" + content + "</HEAD>\n"
  root = ET.fromstring(content)
  return root


def convertLikelihoodToNum(root):
  """Convert entries from likelihood data to absolute number.

  :param root: an elementtree
  :returns: an elementtree with modified entries
  """
  nSignalVec = [int(i) for i in root.find("NSignalEvents").text.split()]
  nBackgroundVec = [int(i) for i in root.find("NBackgroundEvents").text.split()]

  dataChildrenNodes = [node for node in list(root) if node.tag.find("_") > 0]
  for node in dataChildrenNodes:
    nCurrentEnergyBin = int(node.tag.split("_")[-1])
    nBin = nSignalVec[nCurrentEnergyBin] if node.tag.find("Bkg") < 0 else nBackgroundVec[nCurrentEnergyBin]

    subnode = root.find("./" + node.tag + "/BinContents")
    subnode.text = " ".join([str(int(round(float(i) * nBin))) for i in subnode.text.split()])
  return root


def convertNumToLikelihood(root):
  """Convert entries from absolute number to likelihood number.

  :param root: an elementtree
  :returns: an elementtree with modified entries
  """
  nSignalVec = [int(i) for i in root.find("NSignalEvents").text.split()]
  nBackgroundVec = [int(i) for i in root.find("NBackgroundEvents").text.split()]

  dataChildrenNodes = [node for node in list(root) if node.tag.find("_") > 0]
  for node in dataChildrenNodes:
    nCurrentEnergyBin = int(node.tag.split("_")[-1])
    nBin = nSignalVec[nCurrentEnergyBin] if node.tag.find("Bkg") < 0 else nBackgroundVec[nCurrentEnergyBin]

    subnode = root.find("./" + node.tag + "/BinContents")
    # TODO FIXME treat nBin == 0 case. it was in the original script.
    if nBin == 0:
      nBin = 1
    # same precision as we had in python2
    subnode.text = " ".join([('{0:.12g}'.format(float(i) / nBin) if float(i) else '0.0') for i in subnode.text.split()])
  return root


def addTwoNode(initialRoot, finalRoot, tag):
  """Add entries of a tag of two trees.

  :param initialRoot: an elementtree
  :param finalRoot: an elementtree to be modified
  :param tag: the tag to be operated
  :returns: an elementtree with modified entries
  """
  nVecFinal = [int(i) for i in finalRoot.find(tag).text.split()]
  nVecInitial = [int(i) for i in initialRoot.find(tag).text.split()]
  nVecSim = [sum(x) for x in zip(nVecFinal, nVecInitial)]
  finalRoot.find(tag).text = " ".join(str(i) for i in nVecSim)
  return finalRoot


def mergeLikelihoods(inputFiles, outputFile):
  """Change many pandora photon likelihood files into a single likelihood file.

  This is the the main method.
  :param list str inputFiles: a list of strings with names of input files
  :param str outputFile: a string of output file
  """
  finalRoot = None
  for inputFile in inputFiles:
    root = convertLikelihoodToNum(getFixedXMLTree(inputFile))

    if (finalRoot is None):
      finalRoot = root
    else:
      finalRoot = addTwoNode(root, finalRoot, "NSignalEvents")
      finalRoot = addTwoNode(root, finalRoot, "NBackgroundEvents")

      dataChildrenNodes = [node for node in list(root) if node.tag.find("_") > 0]
      for node in dataChildrenNodes:
        finalRoot = addTwoNode(root, finalRoot, "./" + node.tag + "/BinContents")
  try:
    finalRoot = convertNumToLikelihood(finalRoot)
  except ZeroDivisionError as e:
    return S_ERROR('Division by zero. Not enough events used for photon training - some energy bins have 0 bkg/signal'
                   ' events. Error_msg: %s' % e)

  # fix xml; Strip the top element to revert the change
  finalContent = ET.tostring(finalRoot).decode().strip('</HEAD>').strip('<HEAD>').lstrip()
  with open(outputFile, "w") as f:
    f.write(finalContent)
  return S_OK()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Utilities to deal with the normalisation of LFNs for various occasions.

:author: sailer
"""

from __future__ import absolute_import
__RCSID__ = "$Id$"

# use posixpath because it always uses "/"
import posixpath
from DIRAC import gLogger

LOG = gLogger.getSubLogger(__name__)


def joinPathForMetaData(*args):
  """Returns path expected by MetaDataDictionaries, always ends with a slash When paths are used for metadata, the ending "/" will be rstripped."""
  # if there is a lone slash in this list then the end result is only "/" so we remove them
  cleanedEntries = tuple(ent for ent in args if ent != "/")
  LOG.debug("After cleaning", cleanedEntries)
  finalPath = ""
  for entry in cleanedEntries:
    LOG.debug("This entry", entry)
    finalPath = posixpath.join(finalPath, entry)
  LOG.debug("After Joining", finalPath)
  finalPath = posixpath.normpath(finalPath)
  LOG.debug("After norming", finalPath)
  finalPath = finalPath + "/"
  LOG.verbose("Final Path ", finalPath)
  return finalPath


def cleanUpLFNPath(lfn):
  """Normalise LFNs and remove 'LFN:' prefix."""
  LOG.debug("LFN before Cleanup", lfn)
  lfn = posixpath.normpath(lfn)
  if lfn.lower().startswith('lfn'):
    LOG.debug("LFN starts with lfn:'")
    lfn = lfn[4:]
  LOG.verbose("LFN after Cleanup", lfn)
  return lfn



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Production Job class. Used to define new productions.

Mostly similar to :mod:`~ILCDIRAC.Interfaces.API.NewInterface.UserJob`, but
cannot be (and should not be) used like the
:mod:`~ILCDIRAC.Interfaces.API.NewInterface.UserJob` class.

:author: Stephane Poss
:author: Remi Ete
:author: Ching Bon Lam
"""

from __future__ import print_function
from __future__ import absolute_import
import os
import shutil

from collections import defaultdict
from decimal import Decimal

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities.ReturnValues import returnSingleResult
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getVOForGroup
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.Core.Workflow.Module import ModuleDefinition
from DIRAC.Core.Workflow.Step import StepDefinition
from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

from ILCDIRAC.ILCTransformationSystem.Client.Transformation import Transformation
from ILCDIRAC.Interfaces.API.NewInterface.Job import Job
from ILCDIRAC.Interfaces.Utilities import JobHelpers
import six

__RCSID__ = "$Id$"

LOG = gLogger.getSubLogger(__name__)


class ProductionJob(Job):  # pylint: disable=too-many-public-methods, too-many-instance-attributes
  """Production job class.

  Suitable for CLIC studies. Need to sub class and overload for other clients.
  """

  def __init__(self, script=None):
    super(ProductionJob, self).__init__(script)
    self.prodVersion = __RCSID__
    self.dryrun = False
    self.created = False
    self.checked = False
    self.call_finalization = False
    self.finalsdict = {}
    self.transfid = 0
    self.type = 'Production'
    self.csSection = '/Production/Defaults'
    self.ops = Operations()
    self.fc = FileCatalogClient()
    self.trc = TransformationClient()
    self.defaultProdID = '12345'
    self.defaultProdJobID = '12345'
    self.jobFileGroupSize = 1
    self.nbtasks = 1
    self.slicesize = 0
    self.basename = ''
    self.basepath = self.ops.getValue('/Production/CLIC/BasePath', '/ilc/prod/clic/')
    self.evttype = ''
    self.datatype = ''
    self.energycat = ''
    self.detector = ''
    self.currtrans = None
    self.description = ''
    self.generator = None

    self.finalpaths = []
    self.finalMetaDict = defaultdict(dict)
    self.prodMetaDict = {}
    self.finalMetaDictNonSearch = {}
    self.metadict_external = {}
    self.outputStorage = ''

    self.proxyinfo = getProxyInfo()
    self.vo = getVOForGroup(self.proxyinfo.get('Value', {}).get('group', 'ilc_user'))

    self.inputdataquery = False
    self.inputBKSelection = {}
    self.plugin = 'Standard'
    self.prodGroup = ''

    self.prodTypes = ['MCGeneration', 'MCSimulation', 'Test', 'MCReconstruction',
                      'MCReconstruction_Overlay', 'Merge', 'Split',
                      'MCGeneration_ILD',
                      'MCSimulation_ILD',
                      'MCReconstruction_ILD',
                      'MCReconstruction_Overlay_ILD',
                      'Split_ILD'
                     ]
    self.prodparameters = {}
    self.prodparameters['NbInputFiles'] = 1
    self.prodparameters['nbevts'] = 0
    #self.prodparameters["SWPackages"] = ''
    self._addParameter(self.workflow, "IS_PROD", 'JDL', True, "This job is a production job")
    if not script:
      self.__setDefaults()

    self._recBasePaths = {}
    self.maxFCFoldersToCheck = 100000

  #############################################################################
  def __setDefaults(self):
    """Sets some default parameters."""
    self.setPlatform(self.ops.getValue('%s/Platform' % (self.csSection), 'x86_64-slc5-gcc43-opt'))
    self.setCPUTime('300000')
    self.setLogLevel('verbose')
    self.setJobGroup('@{PRODUCTION_ID}')

    # version control
    self._setParameter('productionVersion', 'string', self.prodVersion, 'ProdAPIVersion')

    # General workflow parameters
    self._setParameter('PRODUCTION_ID', 'string', self.defaultProdID.zfill(8), 'ProductionID')
    self._setParameter('JOB_ID', 'string', self.defaultProdJobID.zfill(8), 'ProductionJobID')
    self._setParameter('Priority', 'JDL', '1', 'Priority')
    self._setParameter('emailAddress', 'string', 'ilcdirac-support@cern.ch', 'CrashEmailAddress')

  def _setParameter(self, name, parameterType, parameterValue, description):
    """Set parameters checking in CS in case some defaults need to be changed."""
    if self.ops.getValue('%s/%s' % (self.csSection, name), ''):
      LOG.debug('Setting %s from CS defaults = %s' % (name, self.ops.getValue('%s/%s' % (self.csSection, name))))
      self._addParameter(self.workflow, name, parameterType, self.ops.getValue('%s/%s' % (self.csSection, name),
                                                                               'default'), description)
    else:
      LOG.debug('Setting parameter %s = %s' % (name, parameterValue))
      self._addParameter(self.workflow, name, parameterType, parameterValue, description)

  def setConfig(self, version):
    """Define the Configuration package to obtain."""
    appName = 'ILDConfig'
    self._addSoftware(appName.lower(), version)
    self.prodparameters['ILDConfigVersion'] = version
    self._addParameter(self.workflow, 'ILDConfigPackage', 'JDL', appName + version, 'ILDConfig package')
    return S_OK()

  def setConfigPackage(self, appName, version):
    """Define the config package to obtain.

    Adds Config package to workflow execution, and sets production parameter.
    See :func:`ILCDIRAC.Interfaces.API.NewInterface.Job.Job.setConfigPackage`

    :param str appName: name of the ConfigPackage, e.g. 'ClicConfig'
    :param str version: version of the ConfigPackage
    """
    super(ProductionJob, self).setConfigPackage(appName, version)
    self.prodparameters[appName + 'Version'] = version
    return S_OK()

  def setDryRun(self, run):
    """In case one wants to get all the info as if the prod was being submitted."""
    self.dryrun = run

  #############################################################################
  def setProdGroup(self, group):
    """Sets a user defined tag for the production as appears on the monitoring page."""
    self.prodGroup = group
  #############################################################################

  def setProdPlugin(self, plugin):
    """Sets the plugin to be used to creating the production jobs."""
    self.plugin = plugin

  #############################################################################
  def setJobFileGroupSize(self, files):
    """Sets the number of files to be input to each job created."""
    if self.checked:
      return self._reportError("This input is needed at the beginning of the production definition: it is \
      needed for total number of evts.")
    self.jobFileGroupSize = files
    self.prodparameters['NbInputFiles'] = files

  def setNbEvtsPerSlice(self, nbevts):
    """Define the number of events in a slice."""
    self.slicesize = nbevts

  #############################################################################
  def setProdType(self, prodType):
    """Set prod type."""
    if prodType not in self.prodTypes:
      raise TypeError('Prod must be one of %s' % (', '.join(self.prodTypes)))
    self.setType(prodType)
  #############################################################################

  def setWorkflowName(self, name):
    """Set workflow name."""
    self.workflow.setName(name)
    self.name = name

  #############################################################################
  def setWorkflowDescription(self, desc):
    """Set workflow name."""
    self.workflow.setDescription(desc)

  #############################################################################
  def createWorkflow(self):
    """Create XML for local testing."""
    name = '%s.xml' % self.name
    if os.path.exists(name):
      shutil.move(name, '%s.backup' % name)
    self.workflow.toXMLFile(name)

  #############################################################################
  def setOutputSE(self, outputse):
    """Define where the output file(s) will go."""
    self.outputStorage = outputse
    return S_OK()

  #############################################################################
  def setInputDataQuery(self, metadata):
    """Define the input data query needed."""

    retMetaKey = self._checkMetaKeys(list(metadata.keys()))
    if not retMetaKey['OK']:
      return retMetaKey
    LOG.notice('InputMetaQuery: %s' % metadata)

    if "ProdID" not in metadata:
      return self._reportError("Input metadata dictionary must contain at least a key 'ProdID' as reference")
    retDirs = self._checkFindDirectories(metadata)
    if not retDirs['OK']:
      return retDirs
    dirs = list(retDirs['Value'].values())
    for mdir in dirs[:self.maxFCFoldersToCheck]:
      LOG.notice("Directory: %s" % mdir)
      res = self.fc.getDirectoryUserMetadata(mdir)
      if not res['OK']:
        return self._reportError("Error looking up the catalog for directory metadata")
      compatmeta = res['Value']
      compatmeta.update(metadata)

    if 'EvtType' in compatmeta:
      self.evttype = JobHelpers.getValue(compatmeta['EvtType'], str, six.string_types)
    else:
      return self._reportError("EvtType is not in the metadata, it has to be!")

    if 'NumberOfEvents' in compatmeta:
      self.nbevts = JobHelpers.getValue(compatmeta['NumberOfEvents'], int, None)

    self.basename = self.evttype

    self.generator = compatmeta.get('Generator', None)

    LOG.notice("MetaData: %s" % compatmeta)
    LOG.notice("MetaData: %s" % metadata)
    if "Energy" in compatmeta:
      self.energycat = JobHelpers.getValue(compatmeta["Energy"], str, (int, int, six.string_types))

    if self.energycat.count("tev"):
      self.energy = Decimal("1000.") * Decimal(self.energycat.split("tev")[0])
    elif self.energycat.count("gev"):
      self.energy = Decimal("1.") * Decimal(self.energycat.split("gev")[0])
    else:
      self.energy = Decimal("1.") * Decimal(self.energycat)
    gendata = False
    if 'Datatype' in compatmeta:
      self.datatype = JobHelpers.getValue(compatmeta['Datatype'], str, six.string_types)
      if self.datatype == 'gen':
        gendata = True
    if "DetectorType" in compatmeta and not gendata:
      self.detector = JobHelpers.getValue(compatmeta["DetectorType"], str, six.string_types)
    self.inputBKSelection = metadata
    self.inputdataquery = True

    self.prodparameters['nbevts'] = self.nbevts
    self.prodparameters["FCInputQuery"] = self.inputBKSelection

    return S_OK()

  def setDescription(self, desc):
    """Set the production's description.

    :param str desc: Description
    """
    self.description = desc
    return S_OK()

  def getBasePath(self):
    """Return the base path.

    Updated by :any:`setInputDataQuery`.
    """
    return self.basepath

  def addFinalization(self, uploadData=False, registerData=False, uploadLog=False, sendFailover=False):
    """Add finalization step.

    :param bool uploadData: Upload or not the data to the storage
    :param bool uploadLog: Upload log file to storage (currently only available for admins, thus add them to OutputSandbox)
    :param bool sendFailover: Send Failover requests, and declare files as processed or unused in transfDB
    :param bool registerData: Register data in the file catalog
    """
    # TODO: Do the registration only once, instead of once for each job

    self.call_finalization = True
    self.finalsdict['uploadData'] = uploadData
    self.finalsdict['registerData'] = registerData
    self.finalsdict['uploadLog'] = uploadLog
    self.finalsdict['sendFailover'] = sendFailover

  def _addRealFinalization(self):
    """This is called at creation: now that the workflow is created at the last minute, we need to add this also at the last minute."""
    importLine = 'from ILCDIRAC.Workflow.Modules.<MODULE> import <MODULE>'

    dataUpload = ModuleDefinition('UploadOutputData')
    dataUpload.setDescription('Uploads the output data')
    self._addParameter(dataUpload, 'enable', 'bool', False, 'EnableFlag')
    body = importLine.replace('<MODULE>', 'UploadOutputData')
    dataUpload.setBody(body)

    failoverRequest = ModuleDefinition('FailoverRequest')
    failoverRequest.setDescription('Sends any failover requests')
    self._addParameter(failoverRequest, 'enable', 'bool', False, 'EnableFlag')
    body = importLine.replace('<MODULE>', 'FailoverRequest')
    failoverRequest.setBody(body)

    registerdata = ModuleDefinition('RegisterOutputData')
    registerdata.setDescription('Module to add in the metadata catalog the relevant info about the files')
    self._addParameter(registerdata, 'enable', 'bool', False, 'EnableFlag')
    body = importLine.replace('<MODULE>', 'RegisterOutputData')
    registerdata.setBody(body)

    logUpload = ModuleDefinition('UploadLogFile')
    logUpload.setDescription('Uploads the output log files')
    self._addParameter(logUpload, 'enable', 'bool', False, 'EnableFlag')
    body = importLine.replace('<MODULE>', 'UploadLogFile')
    logUpload.setBody(body)

    errorReport = ModuleDefinition('ReportErrors')
    errorReport.setDescription('Reports errors at the end')
    body = importLine.replace('<MODULE>', 'ReportErrors')
    errorReport.setBody(body)

    finalization = StepDefinition('Job_Finalization')
    finalization.addModule(dataUpload)
    up = finalization.createModuleInstance('UploadOutputData', 'dataUpload')
    up.setValue("enable", self.finalsdict['uploadData'])

    finalization.addModule(registerdata)
    ro = finalization.createModuleInstance('RegisterOutputData', 'RegisterOutputData')
    ro.setValue("enable", self.finalsdict['registerData'])

    finalization.addModule(logUpload)
    ul = finalization.createModuleInstance('UploadLogFile', 'logUpload')
    ul.setValue("enable", self.finalsdict['uploadLog'])

    finalization.addModule(failoverRequest)
    fr = finalization.createModuleInstance('FailoverRequest', 'failoverRequest')
    fr.setValue("enable", self.finalsdict['sendFailover'])

    finalization.addModule(errorReport)
    fr = finalization.createModuleInstance('ReportErrors', 'reportErrors')

    self.workflow.addStep(finalization)
    self.workflow.createStepInstance('Job_Finalization', 'finalization')

    return S_OK()

  def createProduction(self, name=None):
    """Create production.

    :return: S_OK with the created transformation, or S_ERROR
    """
    if not self.dryrun:
      if not self.proxyinfo['OK']:
        return S_ERROR('Not allowed to create production, you need a production proxy.')
      if 'groupProperties' not in self.proxyinfo['Value']:
        return S_ERROR('Could not determine groupProperties, you do not have the right proxy.')
      groupProperties = self.proxyinfo['Value']['groupProperties']
      if 'ProductionManagement' not in groupProperties:
        return S_ERROR('Not allowed to create production, you need a production proxy.')

    if self.created:
      return S_ERROR("Production already created.")

    # We need to add the applications to the workflow
    res = self._addToWorkflow()
    if not res['OK']:
      return res
    if self.call_finalization:
      self._addRealFinalization()

    workflowName = self.workflow.getName()
    fileName = '%s.xml' % workflowName
    LOG.verbose('Workflow XML file name is:', '%s' % fileName)
    try:
      self.createWorkflow()
    except Exception as x:
      LOG.error("Exception creating workflow", repr(x))
      return S_ERROR('Could not create workflow')
    with open(fileName, 'r') as oFile:
      workflowXML = oFile.read()
    if not name:
      name = workflowName

    res = self.trc.getTransformationStats(name)
    if res['OK']:
      return self._reportError("Transformation with name %s already exists! Cannot proceed." % name)

    # Create Tranformation
    Trans = Transformation()
    Trans.setTransformationName(name)
    Trans.setDescription(self.description)
    Trans.setLongDescription(self.description)
    Trans.setType(self.type)
    self.prodparameters['JobType'] = self.type
    Trans.setPlugin(self.plugin)
    if self.inputdataquery:
      Trans.setGroupSize(self.jobFileGroupSize)
    Trans.setTransformationGroup(self.prodGroup)
    Trans.setBody(workflowXML)
    if not self.slicesize:
      Trans.setEventsPerTask(self.jobFileGroupSize * self.nbevts)
    else:
      Trans.setEventsPerTask(self.slicesize)
    self.currtrans = Trans
    if self.inputBKSelection:
      Trans.setInputMetaQuery(self.inputBKSelection)
    if self.dryrun:
      LOG.notice('Would create prod called', name)
      self.transfid = 12345
    else:
      res = Trans.addTransformation()
      if not res['OK']:
        LOG.error(res['Message'])
        return res
      self.transfid = Trans.getTransformationID()['Value']

    if not self.dryrun:
      Trans.setAgentType("Automatic")
      Trans.setStatus("Active")

    finals = []
    for finalpaths in self.finalpaths:
      finalpaths = finalpaths.rstrip("/")
      finalpaths += "/" + str(self.transfid).zfill(8)
      finals.append(finalpaths)
      self.finalMetaDict[finalpaths].update({"ProdID": self.transfid})
      self.finalMetaDict[finalpaths].update(self.prodMetaDict)
      # if 'ILDConfigVersion' in self.prodparameters:
      #   self.finalMetaDict[finalpaths].update({"ILDConfig":self.prodparameters['ILDConfigVersion']})

      if self.nbevts:
        self.finalMetaDict[finalpaths].update({'NumberOfEvents': self.jobFileGroupSize * self.nbevts})
    self.finalpaths = finals
    self.created = True

    return S_OK(Trans)

  def setNbOfTasks(self, nbtasks):
    """Define the number of tasks you want.

    Useful for generation jobs.
    """
    if not self.currtrans:
      LOG.error("Not transformation defined earlier")
      return S_ERROR("No transformation defined")
    if self.inputBKSelection and self.plugin not in ['Limited', 'SlicedLimited']:
      LOG.error('Metadata selection activated, should not specify the number of jobs')
      return S_ERROR()
    self.nbtasks = nbtasks
    self.currtrans.setMaxNumberOfTasks(self.nbtasks)  # pylint: disable=E1101
    return S_OK()

  def addMetadataToFinalFiles(self, metadict):
    """Add additionnal non-query metadata."""
    self.metadict_external = metadict

    return S_OK()

  def finalizeProd(self, prodid=None, prodinfo=None):
    """Finalize definition: submit to Transformation service and register metadata."""
    currtrans = 0
    if self.currtrans:
      if not self.dryrun:
        currtrans = self.currtrans.getTransformationID()['Value']  # pylint: disable=E1101
      else:
        currtrans = 12345
    if prodid:
      currtrans = prodid
    if not currtrans:
      LOG.error("Not transformation defined earlier")
      return S_ERROR("No transformation defined")
    if prodinfo:
      self.prodparameters = prodinfo

    info = []
    info.append('%s Production %s has following parameters:\n' % (self.prodparameters['JobType'], currtrans))
    if "Process" in self.prodparameters:
      info.append('- Process %s' % self.prodparameters['Process'])
    if "Energy" in self.prodparameters:
      info.append('- Energy %s GeV' % self.prodparameters["Energy"])

    if not self.slicesize:
      self.prodparameters['nbevts'] = self.jobFileGroupSize * self.nbevts
    else:
      self.prodparameters['nbevts'] = self.slicesize
    if self.prodparameters['nbevts']:
      info.append("- %s events per job" % (self.prodparameters['nbevts']))
    if self.prodparameters.get('lumi', False):
      info.append('    corresponding to a luminosity %s fb' % (self.prodparameters['lumi']
                                                               * self.prodparameters['NbInputFiles']))
    if 'FCInputQuery' in self.prodparameters:
      info.append('Using InputDataQuery :')
      for key, val in self.prodparameters['FCInputQuery'].items():
        info.append('    %s = %s' % (key, val))
    if "SWPackages" in self.prodparameters:
      info.append('- SW packages %s' % self.prodparameters["SWPackages"])
    if "SoftwareTag" in self.prodparameters:
      info.append('- SW tags %s' % self.prodparameters["SoftwareTag"])
    if "ILDConfigVersion" in self.prodparameters:
      info.append('- ILDConfig %s' % self.prodparameters['ILDConfigVersion'])

    if 'ClicConfigVersion' in self.prodparameters:
      info.append('- ClicConfig %s' % self.prodparameters['ClicConfigVersion'])

    if 'extraCLIArguments' in self.prodparameters:
      info.append('- ExtraCLIArguments %s' % self.prodparameters['extraCLIArguments'])

    # as this is the very last call all applications are registered, so all software packages are known
    # add them the the metadata registration
    for finalpath in self.finalpaths:
      if finalpath not in self.finalMetaDictNonSearch:
        self.finalMetaDictNonSearch[finalpath] = {}
      if "SWPackages" in self.prodparameters:
        self.finalMetaDictNonSearch[finalpath]["SWPackages"] = self.prodparameters["SWPackages"]

      if self.metadict_external:
        self.finalMetaDictNonSearch[finalpath].update(self.metadict_external)

    info.append('- Registered metadata: ')
    for path, metadata in sorted(self.finalMetaDict.items()):
      info.append('    %s = %s' % (path, metadata))
    info.append('- Registered non searchable metadata: ')
    for path, metadata in sorted(self.finalMetaDictNonSearch.items()):
      info.append('    %s = %s' % (path, metadata))

    infoString = '\n'.join(info)
    self.prodparameters['DetailedInfo'] = infoString

    for name, val in self.prodparameters.items():
      result = self._setProdParameter(currtrans, name, val)
      if not result['OK']:
        LOG.error(result['Message'])

    res = self._registerMetadata()
    if not res['OK']:
      LOG.error('Could not register the following directories:', res['Message'])
      return res
    return S_OK()

  def _createDirectory(self, path, failed, mode=0o775):
    """Create the directory at path if it does not exist.

    :param str path: path to check
    :param list failed: list of failed paths
    :param int mode: mode to set for directory
    """
    exists = returnSingleResult(self.fc.isDirectory(path))
    if exists['OK'] and exists['Value']:
      LOG.verbose('Directory already exists:', path)
      return S_OK()
    result = returnSingleResult(self.fc.createDirectory(path))
    if not result['OK']:
      LOG.error('Failed to create directory:', '%s: %s' % (path, result['Message']))
      failed[path].append(result['Message'])
      return S_ERROR()
    LOG.verbose('Successfully created directory:', path)
    res = self.fc.changePathMode({path: mode}, False)
    if not res['OK']:
      LOG.error(res['Message'])
      failed[path].append(res['Message'])
      return S_ERROR()
    LOG.verbose('Successfully changed mode:', path)
    return S_OK()

  def _checkMetadata(self, path, metaCopy):
    """Get existing metadata, if it is the same do not set it again, otherwise return error."""
    existingMetadata = self.fc.getDirectoryUserMetadata(path.rstrip('/'))
    if not existingMetadata['OK']:
      return S_OK()
    failure = False
    for key, value in existingMetadata['Value'].items():
      if key in metaCopy and metaCopy[key] != value:
        LOG.error('Metadata values for folder %s disagree for key %s: Existing(%r), new(%r)' %
                  (path, key, value, metaCopy[key]))
        failure = True
      elif key in metaCopy and metaCopy[key] == value:
        LOG.verbose('Meta entry is unchanged', '%s = %s' % (key, value))
        metaCopy.pop(key, None)
    if failure:
      return S_ERROR('Error when setting new metadata, already existing metadata disagrees!')
    return S_OK()

  def _registerMetadata(self):
    """Set metadata for given folders.

    Register path and metadata before the production actually runs. This allows for the definition
    of the full chain in 1 go.
    """
    prevent_registration = self.ops.getValue('Production/PreventMetadataRegistration', False)

    if self.dryrun or prevent_registration:
      LOG.notice('Would have created and registered the following\n',
                 '\n '.join([' * %s: %s' % (fPath, val) for fPath, val in sorted(self.finalMetaDict.items())]))
      LOG.notice('Would have set this as non searchable metadata', str(self.finalMetaDictNonSearch))
      return S_OK()

    failed = defaultdict(list)
    for path, meta in sorted(self.finalMetaDict.items()):
      res = self._createDirectory(path, failed)
      if not res['OK']:
        continue
      LOG.verbose('Checking to set metadata:', meta)
      metaCopy = dict(meta)
      res = self._checkMetadata(path, metaCopy)
      if not res['OK']:
        return res
      if not metaCopy:
        LOG.verbose('No new metadata to set')
        continue

      LOG.notice('Setting metadata information: ', '%s: %s' % (path, metaCopy))
      result = self.fc.setMetadata(path.rstrip('/'), metaCopy)
      if not result['OK']:
        LOG.error('Could not preset metadata', str(metaCopy))
        LOG.error('Could not preset metadata', result['Message'])
        failed[path].append(result['Message'])

    for path, meta in sorted(self.finalMetaDictNonSearch.items()):
      res = self._createDirectory(path, failed)
      if not res['OK']:
        continue
      LOG.notice('Setting non searchable metadata information: ', '%s: %s' % (path, meta))
      result = self.fc.setMetadata(path.rstrip('/'), meta)
      if not result['OK']:
        LOG.error('Could not preset non searchable metadata', str(meta))
        LOG.error('Could not preset non searchable metadata', result['Message'])
        failed[path].append(result['Message'])

    if failed:
      return S_ERROR('Failed to register some metadata: %s' % dict(failed))
    return S_OK()

  def getMetadata(self):
    """Return the corresponding metadata of the last step."""
    metadict = {}
    for meta in self.finalMetaDict.values():
      metadict.update(meta)
    if 'NumberOfEvents' in metadict:
      del metadict['NumberOfEvents']  # As this is not supposed to be a searchable thing
    return metadict

  def _setProdParameter(self, prodID, pname, pvalue):
    """Set a production parameter."""
    if isinstance(pvalue, list):
      pvalue = '\n'.join(pvalue)

    if isinstance(pvalue, six.integer_types):
      pvalue = str(pvalue)
    if not self.dryrun:
      result = self.trc.setTransformationParameter(int(prodID), str(pname), str(pvalue))
      if not result['OK']:
        LOG.error('Problem setting parameter %s for production %s and value:\n%s' % (prodID, pname, pvalue))
    else:
      LOG.notice('Adding %s=%s to transformation' % (str(pname), str(pvalue)))
      result = S_OK()
    return result

  def _jobSpecificParams(self, application):
    """For production additional checks are needed: ask the user."""

    if self.created:
      return S_ERROR("The production was created, you cannot add new applications to the job.")

    if not application.logFile:
      logf = application.appname + "_" + application.version + "_@{STEP_ID}.log"
      res = application.setLogFile(logf)
      if not res['OK']:
        return res

      # in fact a bit more tricky as the log files have the prodID and jobID in them

    # Retrieve from the application the essential info to build the prod info.
    if not self.nbevts and not self.slicesize:
      self.nbevts = application.numberOfEvents
      if not self.nbevts:
        return S_ERROR("Number of events to process is not defined.")
    elif not application.numberOfEvents:
      if not self.slicesize:
        res = application.setNumberOfEvents(self.jobFileGroupSize * self.nbevts)
      else:
        res = application.setNumberOfEvents(self.slicesize)
      if not res['OK']:
        return res

    if application.numberOfEvents > 0 and (
        self.jobFileGroupSize
        * self.nbevts > application.numberOfEvents or self.slicesize > application.numberOfEvents):
      self.nbevts = application.numberOfEvents

    if not self.energy:
      if application.energy:
        self.energy = Decimal((("%1.2f" % float(application.energy)).rstrip('0').rstrip('.')))
      else:
        return S_ERROR("Could not find the energy defined, it is needed for the production definition.")
    elif not application.energy:
      res = application.setEnergy(float(self.energy))
      if not res['OK']:
        return res
    if self.energy:
      self._setParameter("Energy", "float", float(self.energy), "Energy used")
      self.prodparameters["Energy"] = float(self.energy)

    if not self.evttype:
      if hasattr(application, 'eventType'):
        self.evttype = application.eventType
      else:
        return S_ERROR("Event type not found nor specified, it's mandatory for the production paths.")
      self.prodparameters['Process'] = self.evttype

    if not self.outputStorage:
      return S_ERROR("You need to specify the Output storage element")

    curpackage = "%s.%s" % (application.appname, application.version)
    if "SWPackages" in self.prodparameters:
      if not self.prodparameters["SWPackages"].count(curpackage):
        self.prodparameters["SWPackages"] += ";%s" % (curpackage)
    else:
      self.prodparameters["SWPackages"] = "%s" % (curpackage)

    if not application.accountInProduction:
      res = self._updateProdParameters(application)
      if not res['OK']:
        return res
      self.checked = True

      return S_OK()

    res = application.setOutputSE(self.outputStorage)
    if not res['OK']:
      return res

    energypath = self.getEnergyPath()

    if not self.basename:
      self.basename = self.evttype

    evttypepath = ''
    if not self.evttype[-1] == '/':
      evttypepath = self.evttype + '/'

    path = self.basepath

    # Need to resolve file names and paths
    if self.energy:
      self.finalMetaDict[self.basepath + energypath] = {"Energy": str(self.energy)}

    currentPathList = [self.basepath, energypath, evttypepath]
    path = os.path.join(*currentPathList)
    self.finalMetaDict[path] = {"EvtType": self.evttype}
    if self.generator:
      self.prodparameters['Generator'] = self.generator
      currentPathList.append(self.generator)
      generatorPath = os.path.join(*currentPathList)
      self.finalMetaDict[generatorPath] = {'Generator': self.generator}

    if hasattr(application, "setOutputRecFile") and not application.willBeCut:
      detPath = os.path.join(path, application.detectortype)
      self.finalMetaDict[detPath] = {'DetectorType': application.detectortype}
      if application.keepRecFile:
        path = os.path.join(detPath, 'REC')
        self.finalMetaDict[path] = {'Datatype': 'REC'}
        fname = self.basename + '_rec.slcio'
        application.setOutputRecFile(fname, path)
        LOG.notice('Will store the files under', path)
        self.finalpaths.append(path)
      path = os.path.join(detPath, 'DST')
      self.finalMetaDict[path] = {'Datatype': 'DST'}
      fname = self.basename + '_dst.slcio'
      application.setOutputDstFile(fname, path)
      LOG.notice('Will store the files under', path)
      self.finalpaths.append(path)

    elif hasattr(application, "outputFile") and hasattr(application, 'datatype') and not application.outputFile and not application.willBeCut:
      LOG.notice('Adding output meta data for %s' % type(application))
      path = os.path.join(*currentPathList)

      if not application.datatype and self.datatype:
        application.datatype = self.datatype

      # for fcc datatype comes before detctor model
      if self.vo == 'fcc':
        path = os.path.join(path, application.datatype)
        self.finalMetaDict[path] = {'Datatype': application.datatype}

      if hasattr(application, "detectortype"):
        if application.detectortype:
          path = os.path.join(path, application.detectortype)
          self.finalMetaDict[path] = {"DetectorType": application.detectortype}
        elif self.detector:
          path = os.path.join(path, self.detector)
          self.finalMetaDict[path] = {"DetectorType": self.detector}
      if hasattr(application, "detector"):
        if application.detector:
          path = os.path.join(path, application.detector)
          self.finalMetaDict[path] = {"DetectorType": application.detector}

      # datatype comes after detctor model
      if self.vo != 'fcc':
        path = os.path.join(path, application.datatype)
        self.finalMetaDict[path] = {'Datatype': application.datatype}

      LOG.notice('Will store the files under', '%s' % path)
      self.finalpaths.append(path)
      extension = getattr(application, '_extension', 'stdhep')
      if application.datatype in ['SIM', 'REC']:
        extension = 'slcio'
      fname = self.basename + "_%s" % (application.datatype.lower()) + "." + extension
      application.setOutputFile(fname, path)

    res = self._updateProdParameters(application)
    if not res['OK']:
      return res

    self.checked = True

    return S_OK()

  def _updateProdParameters(self, application):
    """Update the prod parameters stored in the production parameters visible from the web."""
    try:
      self.prodparameters.update(application.prodparameters)
    except Exception as x:
      return S_ERROR("Exception: %r" % x)

    if hasattr(application, 'extraCLIArguments') and application.extraCLIArguments:
      self.prodparameters['extraCLIArguments'] = repr(application.extraCLIArguments)

    return S_OK()

  def _jobSpecificModules(self, application, step):
    return application._prodjobmodules(step)

  def getEnergyPath(self):
    """returns the energy path 250gev or 3tev or 1.4tev etc."""
    energy = Decimal(str(self.energy))
    tD = Decimal('1000.0')
    unit = 'gev' if energy < tD else 'tev'
    energy = energy if energy < tD else energy / tD
    energyPath = ("%1.2f" % energy).rstrip('0').rstrip('.')
    energyPath = energyPath + unit + '/'

    LOG.notice('Energy path is: ', energyPath)
    return energyPath

  def _checkMetaKeys(self, metakeys, extendFileMeta=False):
    """check if metadata keys are allowed to be metadata.

    :param list metakeys: metadata keys for production metadata
    :param bool extendFileMeta: also use FileMetaFields for checking meta keys
    :returns: S_OK, S_ERROR
    """

    res = self.fc.getMetadataFields()
    if not res['OK']:
      LOG.error("Could not contact File Catalog")
      return S_ERROR("Could not contact File Catalog")
    metaFCkeys = list(res['Value']['DirectoryMetaFields'].keys())
    if extendFileMeta:
      metaFCkeys.extend(list(res['Value']['FileMetaFields'].keys()))

    for key in metakeys:
      for meta in metaFCkeys:
        if meta != key and meta.lower() == key.lower():
          return self._reportError("Key syntax error %r, should be %r" % (key, meta), name=self.__class__.__name__)
      if key not in metaFCkeys:
        return self._reportError("Key %r not found in metadata keys, allowed are %r" % (key, metaFCkeys))

    return S_OK()

  def _checkFindDirectories(self, metadata):
    """find directories by metadata and check that there are directories found.

    :param dict metadata: metadata dictionary
    :returns: S_OK, S_ERROR
    """

    metaQuery = ' '.join('%s=%s' % (k, v) for k, v in metadata.items())
    LOG.verbose('Looking for folder with', repr(metaQuery))
    res = self.fc.findDirectoriesByMetadata(metadata)
    if not res['OK']:
      return self._reportError("Error looking up the catalog for available directories")
    elif len(res['Value']) < 1:
      return self._reportError('Could not find any directories corresponding to the query issued: %s' % metaQuery)
    for folderId, folder in res['Value'].items():
      if (folderId == 0 or folder == 'None') and not self.dryrun:
        return self._reportError('Could not find any directories corresponding to the query issued: %s' % metaQuery)
    return res

  def setReconstructionBasePaths(self, recPath, dstPath):
    """set the output Base paths for the reconstruction REC and DST files."""
    self._recBasePaths['REC'] = recPath
    self._recBasePaths['DST'] = dstPath



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Interface for a GenericApplication.

Allows one to use a script in an application framework. Can also use a compiled executable as long as
all the libraries and supplemental files are available.

See here for more information about running your own executable: https://twiki.cern.ch/twiki/bin/view/CLIC/IlcdiracUserLibraries
"""
from __future__ import absolute_import
__RCSID__ = "$Id$"
import types
import os

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Workflow.Parameter import Parameter

from ILCDIRAC.Interfaces.API.NewInterface.LCApplication import LCApplication


class GenericApplication(LCApplication):
  """Run a script (python or shell) in an application environment.

  Example:

  >>> ga = GenericApplication()
  >>> ga.setScript("myscript.py")
  >>> ga.setArguments("some command line arguments")
  >>> ga.setDependency({"root":"5.26"})

  In case you also use the :func:`setExtraCLIArguments` method, whatever you put
  in there will be added at the end of the command line, i.e., after the Arguments
  """

  def __init__(self, paramdict=None):
    self.script = None
    self.setupScript = ''
    self.arguments = ''
    self.dependencies = {}
    # The Application init has to come last as if not the passed parameters are overwritten by the defaults.
    super(GenericApplication, self).__init__(paramdict)
    # Those have to come last as the defaults from Application are not right
    self._modulename = "ApplicationScript"
    self.appname = self._modulename
    self._moduledescription = 'An Application script module that can execute any provided script in the given \
    project name and version environment'

  def setScript(self, script):
    """Define script to use.

    :param str script: Script to run. Can be shell or python or compiled executable. Can be local file or LFN.
    """
    self._checkArgs({'script': (str,)})
    if os.path.exists(script) or script.lower().count("lfn:"):
      self.inputSB.append(script)
    self.script = script
    return S_OK()

  def setSetupScript(self, setupScript):
    """Define setup script to use.

    :param str setupScript: Shell script to source. Can be local file or LFN.
    """
    self._checkArgs({'setupScript': (str,)})
    if not setupScript.endswith('.sh'):
      return S_ERROR("The script must be a shell file")

    if os.path.exists(setupScript) or setupScript.lower().count("lfn:"):
      self.inputSB.append(setupScript)
    self.setupScript = setupScript
    return S_OK()

  def setArguments(self, args):
    """ Optional: Define the arguments of the script

    :param str args: Arguments to pass to the command call
    """
    self._checkArgs({'args': (str,)})
    self.arguments = args
    return S_OK()

  def setDependency(self, appdict):
    """Define list of application you need.

    >>> app.setDependency({"mokka":"v0706P08","marlin":"v0111Prod"})

    :param dict appdict: Dictionary of application to use: {"App":"version"}
    """
    # check that dict has proper structure
    self._checkArgs({'appdict': dict})

    self.dependencies.update(appdict)
    return S_OK()

  def _applicationModule(self):
    m1 = self._createModuleDefinition()
    m1.addParameter(Parameter("script", "", "string", "", "", False, False, "Script to execute"))
    m1.addParameter(Parameter("arguments", "", "string", "", "", False, False, "Arguments to pass to the script"))
    m1.addParameter(Parameter("debug", False, "bool", "", "", False, False, "debug mode"))
    m1.addParameter(Parameter("setupScript", "", "string", "", "", False, False, "Script to source"))
    return m1

  def _applicationModuleValues(self, moduleinstance):
    moduleinstance.setValue("script", self.script)
    moduleinstance.setValue("setupScript", self.setupScript)
    moduleinstance.setValue('arguments', self.arguments)
    moduleinstance.setValue('debug', self.debug)

  def _userjobmodules(self, stepdefinition):
    res1 = self._setApplicationModuleAndParameters(stepdefinition)
    res2 = self._setUserJobFinalization(stepdefinition)
    if not res1["OK"] or not res2["OK"]:
      return S_ERROR('userjobmodules failed')
    return S_OK()

  def _prodjobmodules(self, stepdefinition):
    res1 = self._setApplicationModuleAndParameters(stepdefinition)
    res2 = self._setOutputComputeDataList(stepdefinition)
    if not res1["OK"] or not res2["OK"]:
      return S_ERROR('prodjobmodules failed')
    return S_OK()

  def _addParametersToStep(self, stepdefinition):
    res = self._addBaseParameters(stepdefinition)
    if not res["OK"]:
      return S_ERROR("Failed to set base parameters")
    return S_OK()

  def _setStepParametersValues(self, instance):
    self._setBaseStepParametersValues(instance)
    for depn, depv in self.dependencies.items():
      self._job._addSoftware(depn, depv)
    return S_OK()

  def _checkConsistency(self, job=None):
    """Checks that script and dependencies are set."""
    if not self.script:
      return S_ERROR("Script not defined")
    elif not self.script.lower().count("lfn:") and not os.path.exists(self.script):
      return S_ERROR("Specified script is not an LFN and was not found on disk")

    if self.setupScript and not self.setupScript.lower().count("lfn:") and not os.path.exists(self.setupScript):
      return S_ERROR("Specified source script is not an LFN and was not found on disk")

    # if not len(self.dependencies):
    #  return S_ERROR("Dependencies not set: No application to install. If correct you should use job.setExecutable")
    return S_OK()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Run SLICPandora.

:since: Oct 25, 2010
:author: sposs
"""

from __future__ import absolute_import
import os
import six.moves.urllib.request
import six.moves.urllib.parse
import six.moves.urllib.error

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities.Subprocess import shellCall

from ILCDIRAC.Workflow.Modules.ModuleBase import ModuleBase
from ILCDIRAC.Core.Utilities.CombinedSoftwareInstallation import getSoftwareFolder, getEnvironmentScript, unzip_file_into_dir
from ILCDIRAC.Core.Utilities.resolvePathsAndNames import resolveIFpaths
from ILCDIRAC.Core.Utilities.PrepareOptionFiles import getNewLDLibs, getNewPATH
from ILCDIRAC.Core.Utilities.PrepareLibs import removeLibc


__RCSID__ = '$Id$'
LOG = gLogger.getSubLogger(__name__)


class SLICPandoraAnalysis (ModuleBase):
  """Run SLIC Pandora."""

  def __init__(self):
    super(SLICPandoraAnalysis, self).__init__()
    self.STEP_NUMBER = ''
    self.result = S_ERROR()
    self.applicationName = 'SLICPandora'
    self.pandorasettings = ""
    self.detectorxml = ""
    self.startFrom = 0
    self.eventstring = ['>>>>>> EVENT']

  def applicationSpecificInputs(self):

    if not self.pandorasettings:
      self.pandorasettings = "PandoraSettings.xml"

    if not len(self.InputFile) and len(self.InputData):
      for files in self.InputData:
        if files.lower().find(".slcio") > -1:
          self.InputFile.append(files)

    return S_OK('Parameters resolved')

  def runIt(self):
    """Called from ModuleBase."""
    self.result = S_OK()
    if not self.platform:
      self.result = S_ERROR('No ILC platform selected')
    elif not self.applicationLog:
      self.result = S_ERROR('No Log file provided')
    if not self.result['OK']:
      LOG.error("Failed to resolve input parameters:", self.result["Message"])
      return self.result

    if not self.workflowStatus['OK'] or not self.stepStatus['OK']:
      LOG.verbose('Workflow status = %s, step status = %s' % (self.workflowStatus['OK'], self.stepStatus['OK']))
      return S_OK('SLIC Pandora should not proceed as previous step did not end properly')

    # Get the env script
    res = getEnvironmentScript(self.platform, self.applicationName, self.applicationVersion, self.getEnvScript)
    if not res['OK']:
      LOG.error("Failed to get the environment script:", res["Message"])
      return res
    env_script_path = res["Value"]

    res = resolveIFpaths(self.InputFile)
    if not res['OK']:
      LOG.error('Could not find input files')
      self.setApplicationStatus('SLICPandora: missing slcio file')
      return S_ERROR('Missing slcio file!')
    runonslcio = res['Value'][0]

    if not self.detectorxml.count(".xml") or not os.path.exists(os.path.basename(self.detectorxml)):
      detmodel = self.detectorxml.replace("_pandora.xml", "")
      if os.path.exists(detmodel + ".zip"):
        try:
          unzip_file_into_dir(open(detmodel + ".zip", 'rb'), os.getcwd())
        except (RuntimeError, OSError) as err:
          LOG.error("Exception when unpacking detectormodel zip file:", str(err))
          os.unlink(detmodel + ".zip")
      if not os.path.exists(detmodel + ".zip"):
        # retrieve detector model from web
        detector_urls = self.ops.getValue('/SLICweb/SLICDetectorModels', [''])
        if len(detector_urls[0]) < 1:
          LOG.error('Could not find in CS the URL for detector model')
          return S_ERROR('Could not find in CS the URL for detector model')

        for detector_url in detector_urls:
          try:
            six.moves.urllib.request.urlretrieve("%s%s" % (detector_url, detmodel + ".zip"), detmodel + ".zip")
          except IOError:
            LOG.error("Download of detector model failed")
            continue
          try:
            unzip_file_into_dir(open(detmodel + ".zip", 'rb'), os.getcwd())
            break
          except (RuntimeError, OSError) as err:
            LOG.error("Exception for zip file obtained from ", detector_url)
            LOG.error("Exception:", str(err))
            os.unlink(detmodel + ".zip")
            continue
      # if os.path.exists(detmodel): #and os.path.isdir(detmodel):
      self.detectorxml = os.path.join(os.getcwd(), self.detectorxml)
      if not self.detectorxml.endswith('xml'):
        self.detectorxml = self.detectorxml + "_pandora.xml"

    if not os.path.exists(self.detectorxml):
      LOG.error('Detector model xml %s was not found, exiting' % self.detectorxml)
      return S_ERROR('Detector model xml was not found, exiting')

    oldversion = False
    if self.applicationVersion in ['CLIC_CDR', 'CDR1', 'CDR2', 'CDR0', 'V2', 'V3', 'V4']:
      oldversion = True

    scriptName = 'SLICPandora_%s_Run_%s.sh' % (self.applicationVersion, self.STEP_NUMBER)
    if os.path.exists(scriptName):
      os.remove(scriptName)
    script = open(scriptName, 'w')
    script.write('#!/bin/bash \n')
    script.write('#####################################################################\n')
    script.write('# Dynamically generated script to run a production or analysis job. #\n')
    script.write('#####################################################################\n')
    script.write("source %s\n" % env_script_path)
    script.write("declare -x file=./Settings/%s\n" % self.pandorasettings)
    script.write("""
if [ -e "${file}" ]
then
   declare -x PANDORASETTINGS=$file
else
  if [ -d "${PANDORASETTINGSDIR}" ]
  then
    cp $PANDORASETTINGSDIR/*.xml .
    declare -x PANDORASETTINGS=%s
  fi
fi
if [ ! -e "${PANDORASETTINGS}" ]
then
  echo "Missing PandoraSettings file"
  exit 1
fi
""" % self.pandorasettings)
    script.write('echo =============================\n')
    script.write('echo PATH is \n')
    script.write('echo $PATH | tr ":" "\n"  \n')
    script.write('echo ==============\n')
    if os.path.exists("./lib"):
      script.write('declare -x LD_LIBRARY_PATH=./lib:$LD_LIBRARY_PATH\n')
    script.write('echo =============================\n')
    script.write('echo LD_LIBRARY_PATH is \n')
    script.write('echo $LD_LIBRARY_PATH | tr ":" "\n"\n')
    script.write('echo ============================= \n')
    script.write('env | sort >> localEnv.log\n')
    # Now run it
    if oldversion:
      comm = 'PandoraFrontend %s $PANDORASETTINGS %s %s %s' % (self.detectorxml, runonslcio,
                                                               self.OutputFile, str(self.NumberOfEvents))
    else:
      comm = 'PandoraFrontend -g %s -c $PANDORASETTINGS -i %s -o %s -r %s' % (self.detectorxml,
                                                                              runonslcio, self.OutputFile,
                                                                              str(self.NumberOfEvents))
    comm = "%s %s\n" % (comm, self.extraCLIarguments)
    LOG.info("Will run %s" % comm)
    script.write(comm)
    script.write('declare -x appstatus=$?\n')
    # script.write('where\n')
    # script.write('quit\n')
    # script.write('EOF\n')
    script.write('exit $appstatus\n')

    script.close()
    if os.path.exists(self.applicationLog):
      os.remove(self.applicationLog)

    os.chmod(scriptName, 0o755)
    comm = 'sh -c "./%s"' % (scriptName)
    self.setApplicationStatus('SLICPandora %s step %s' % (self.applicationVersion, self.STEP_NUMBER))
    self.stdError = ''
    self.result = shellCall(0, comm, callbackFunction=self.redirectLogOutput,
                            bufferLimit=20971520)
    #self.result = {'OK':True,'Value':(0,'Disabled Execution','')}
    resultTuple = self.result['Value']
    if resultTuple[0]:
      LOG.error("There was an error during the execution")

    if not os.path.exists(self.applicationLog):
      LOG.error("Something went terribly wrong, the log file is not present")
      self.setApplicationStatus('%s failed terribly, you are doomed!' % (self.applicationName))
      if not self.ignoreapperrors:
        return S_ERROR('%s did not produce the expected log' % (self.applicationName))

    logf = open(self.applicationLog, 'r')
    if "Missing PandoraSettings file" in logf.readlines()[-1]:
      LOG.error("Issue with the Pandora Settings file.")
    logf.close()

    status = resultTuple[0]
    # stdOutput = resultTuple[1]
    # stdError = resultTuple[2]
    LOG.info("Status after the application execution is %s" % str(status))

    return self.finalStatusReport(status)
    #############################################################################

  def getEnvScript(self, sysconfig, appname, appversion):
    """Produce the environment file in case CVMFS is not here."""
    env_script_name = "SLICPandora.sh"

    res = getSoftwareFolder(sysconfig, appname, appversion)
    if not res['OK']:
      self.setApplicationStatus('SLICPandora: Could not find neither local area not shared area install')
      return res
    myslicPandoraDir = res['Value']

    # Remove libc lib
    removeLibc(myslicPandoraDir + "/LDLibs")

    # Need to fetch the new LD_LIBRARY_PATH
    new_ld_lib_path = getNewLDLibs(sysconfig, appname, appversion)

    new_path = getNewPATH(sysconfig, appname, appversion)

    prefixpath = ""
    if os.path.exists("PandoraFrontend"):
      prefixpath = "."
    elif os.path.exists("%s/Executable/PandoraFrontend" % myslicPandoraDir):
      prefixpath = "%s/Executable" % myslicPandoraDir
    else:
      return S_ERROR("Missing PandoraFrontend binary")
    with open(env_script_name, "w") as script:
      script.write('#!/bin/sh \n')
      script.write('############################################################\n')
      script.write('# Dynamically generated script to get the SLICPandora env. #\n')
      script.write('############################################################\n')
      script.write("declare -x PATH=%s:$PATH\n" % new_path)
      script.write('declare -x ROOTSYS=%s/ROOT\n' % (myslicPandoraDir))
      script.write('declare -x LD_LIBRARY_PATH=$ROOTSYS/lib:%s/LDLibs:%s\n' % (myslicPandoraDir, new_ld_lib_path))
      script.write('declare -x PANDORASETTINGSDIR=%s/Settings\n' % myslicPandoraDir)
      script.write("declare -x PATH=%s:$PATH\n" % prefixpath)
    os.chmod(env_script_name, 0o755)
    return S_OK(os.path.abspath(env_script_name))



#!/usr/local/env python
#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Test OverlayInput module."""
from __future__ import absolute_import
import inspect
import unittest
from mock import patch, MagicMock as Mock

from DIRAC import S_OK, S_ERROR
from ILCDIRAC.Interfaces.API.NewInterface.Applications import OverlayInput
from Tests.Utilities.GeneralUtils import assertEqualsImproved, assertDiracFailsWith, \
    assertDiracSucceeds

__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.Interfaces.API.NewInterface.Applications.OverlayInput'

# pylint: disable=protected-access,too-many-public-methods


class OverlayInputTestCase(unittest.TestCase):
  """Base class for the OverlayInput test cases."""

  def setUp(self):
    """set up the objects."""
    self.olin = OverlayInput({})

  def test_setters(self):
    assertEqualsImproved((self.olin.machine, self.olin.prodID, self.olin.useEnergyForFileLookup,
                            self.olin.detectorModel, self.olin.backgroundEventType),
                          ('clic_cdr', 0, True, '', ''), self)
    self.assertFalse(self.olin._errorDict)
    self.olin.setMachine('mytestMachine')
    assertDiracSucceeds(self.olin.setProdID(14983), self)
    self.olin.setUseEnergyForFileLookup(False)
    self.olin.setDetectorModel('mytestDetector')
    self.olin.setBackgroundType('myBackgroundParticle.Testme')
    assertEqualsImproved((self.olin.machine, self.olin.prodID, self.olin.useEnergyForFileLookup,
                            self.olin.detectorModel, self.olin.backgroundEventType),
                          ('mytestMachine', 14983, False, 'mytestDetector', 'myBackgroundParticle.Testme'),
                          self)
    self.assertFalse(self.olin._errorDict)

  def test_setters_checks(self):
    self.olin.setMachine({'138rj': True})
    self.olin.setProdID([])
    self.olin.setUseEnergyForFileLookup('False')
    self.olin.setDetectorModel(9024)
    self.olin.setBackgroundType(8914)
    assertEqualsImproved(len(self.olin._errorDict['_checkArgs']), 5, self)

  def test_userjobmodules(self):
    with patch.object(self.olin, '_setApplicationModuleAndParameters', new=Mock(return_value=S_OK())):
      assertDiracSucceeds(self.olin._userjobmodules(Mock()), self)

  def test_userjobmodules_fails(self):
    with patch.object(self.olin, '_setApplicationModuleAndParameters', new=Mock(return_value=S_ERROR('bla'))):
      assertDiracFailsWith(self.olin._userjobmodules(Mock()), 'userjobmodules failed', self)

  def test_prodjobmodules(self):
    with patch.object(self.olin, '_setApplicationModuleAndParameters', new=Mock(return_value=S_OK())):
      assertDiracSucceeds(self.olin._prodjobmodules(Mock()), self)

  def test_prodjobmodules_fails(self):
    with patch.object(self.olin, '_setApplicationModuleAndParameters', new=Mock(return_value=S_ERROR('bla'))):
      assertDiracFailsWith(self.olin._prodjobmodules(Mock()), 'prodjobmodules failed', self)

  def test_addparameterstostep_fails(self):
    with patch.object(self.olin, '_addBaseParameters', new=Mock(return_value=S_ERROR('bla'))):
      assertDiracFailsWith(self.olin._addParametersToStep(Mock()), 'failed to set base param', self)

  def test_addparameterstostep(self):
    with patch.object(self.olin, '_addBaseParameters', new=Mock(return_value=S_OK())):
      assertDiracSucceeds(self.olin._addParametersToStep(Mock()), self)

  def test_checkconsistency(self):
    self.olin.pathToOverlayFiles = ''
    self.olin.bxToOverlay = True
    self.olin.numberOfGGToHadronInteractions = 1345
    self.olin.backgroundEventType = 'myTestBGEvt'
    self.olin._jobtype = 'notUser'
    assertDiracSucceeds(self.olin._checkConsistency(), self)
    assertEqualsImproved(
        (self.olin.prodparameters['detectorModel'], self.olin.prodparameters['BXOverlay'],
         self.olin.prodparameters['GGtoHadInt']),
        ('', True, 1345), self)

  def test_checkconsistency_nofilesinpath(self):
    self.olin.pathToOverlayFiles = '/my/path/overlay.files'
    with patch('DIRAC.Resources.Catalog.FileCatalogClient.FileCatalogClient.findFilesByMetadata', new=Mock(return_value=S_OK([]))) as fcc_mock:
      assertDiracFailsWith(self.olin._checkConsistency(), 'no files in that path', self)
      fcc_mock.assert_called_once_with({}, '/my/path/overlay.files')

  def test_checkconsistency_no_bunchcrossings(self):
    self.olin.pathToOverlayFiles = ''
    self.olin.bxToOverlay = None
    assertDiracFailsWith(self.olin._checkConsistency(), 'number of overlay bunch crossings not defined', self)

  def test_checkconsistency_no_backgroundevts(self):
    self.olin.pathToOverlayFiles = ''
    self.olin.bxToOverlay = True
    self.olin.numberOfGGToHadronInteractions = 0
    assertDiracFailsWith(self.olin._checkConsistency(), 'background events per bunch crossing is not defined',
                          self)

  def test_checkconsistency_no_bgevttype(self):
    self.olin.pathToOverlayFiles = ''
    self.olin.bxToOverlay = True
    self.olin.numberOfGGToHadronInteractions = 1345
    self.olin.backgroundEventType = ''
    assertDiracFailsWith(self.olin._checkConsistency(), 'event type is not defined', self)

  def test_checkconsistency_nosignalevtperjob(self):
    self.olin.pathToOverlayFiles = ''
    self.olin.bxToOverlay = True
    self.olin.numberOfGGToHadronInteractions = 1345
    self.olin.backgroundEventType = 'gghad'
    self.olin._jobtype = 'User'
    self.olin.numberOfSignalEventsPerJob = 0
    assertDiracFailsWith(self.olin._checkConsistency(), 'signal event per job is not defined',
                          self)

  def test_checkfinalconsistency_simple(self):
    self.olin.pathToOverlayFiles = 'some/path'
    assertDiracSucceeds(self.olin._checkFinalConsistency(), self)

  def test_checkfinalconsistency_noenergy(self):
    self.olin.energy = 0
    assertDiracFailsWith(self.olin._checkFinalConsistency(), 'energy must be specified', self)

  def test_checkfinalconsistency_ops_fails(self):
    ops_mock = Mock()
    ops_mock.getSections.return_value = S_ERROR('some_ops_error')
    self.olin._ops = ops_mock
    self.olin.energy = 198
    assertDiracFailsWith(self.olin._checkFinalConsistency(), 'could not resolve the cs path', self)

  def test_checkfinalconsistency_machinemissing(self):
    ops_mock = Mock()
    ops_mock.getSections.return_value = S_OK(['allowed_machine_1', 'other_mach'])
    self.olin._ops = ops_mock
    self.olin.machine = 'myTestMachineVeryRare'
    self.olin.energy = 198
    res = self.olin._checkFinalConsistency()
    assertDiracFailsWith(res,
                         "machine 'mytestmachineveryrare' does not have overlay data", self)

  def test_checkfinalconsistency_energynotfound(self):
    self.olin.energy = 824
    section_dict = {'/Overlay': ['myTestMachineVeryRare'],
                     '/Overlay/myTestMachineVeryRare': ['other_energy_tev', '824tev']}
    ops_mock = Mock()
    ops_mock.getSections.side_effect = lambda path: S_OK(section_dict[path])
    self.olin._ops = ops_mock
    self.olin.machine = 'myTestMachineVeryRare'
    assertDiracFailsWith(self.olin._checkFinalConsistency(),
                          'no overlay files corresponding to 824gev', self)

  def test_checkfinalconsistency_nodetectormodels(self):
    self.olin.energy = 1000.0
    section_dict = {'/Overlay': S_OK(['myTestMachineVeryRare']),
                     '/Overlay/myTestMachineVeryRare': S_OK(['other_energy_tev', '1tev']),
                     '/Overlay/myTestMachineVeryRare/1tev': S_ERROR('bla')}
    ops_mock = Mock()
    ops_mock.getSections.side_effect = lambda path: section_dict[path]
    self.olin._ops = ops_mock
    self.olin.machine = 'myTestMachineVeryRare'
    assertDiracFailsWith(self.olin._checkFinalConsistency(),
                          'could not find the detector models', self)

  def test_checkfinalconsistency_detectormodelnotfound(self):
    self.olin.energy = 981324
    section_dict = {'/Overlay': ['myTestMachineVeryRare'],
                     '/Overlay/myTestMachineVeryRare': ['other_energy_tev', '981.3gev'],
                     '/Overlay/myTestMachineVeryRare/981.3tev': ['some_other_detector_model',
                                                                   'neither_this']}
    ops_mock = Mock()
    ops_mock.getSections.side_effect = lambda path: S_OK(section_dict[path])
    self.olin._ops = ops_mock
    self.olin.machine = 'myTestMachineVeryRare'
    assertDiracFailsWith(self.olin._checkFinalConsistency(),
                          'no overlay files corresponding to 981.3tev', self)

  def test_checkfinalconsistency_nobkg(self):
    self.olin.detectorModel = 'testDetMod'
    self.olin.energy = 981000
    section_dict = {'/Overlay': ['myTestMachineVeryRare'],
                     '/Overlay/myTestMachineVeryRare': ['other_energy_tev', '981tev'],
                     '/Overlay/myTestMachineVeryRare/981tev': ['testDetMod']}
    ops_mock = Mock()
    ops_mock.getSections.side_effect = lambda path: S_OK(section_dict[path])
    self.olin._ops = ops_mock
    self.olin.machine = 'myTestMachineVeryRare'
    with patch.object(inspect.getmodule(OverlayInput), 'allowedBkg', new=Mock(return_value=S_ERROR('bkg_test_err'))):
      assertDiracFailsWith(self.olin._checkFinalConsistency(), 'bkg_test_err', self)

  def test_checkfinalconsistency_negativeprodid(self):
    self.olin.detectorModel = 'testDetMod'
    self.olin.energy = 981000
    section_dict = {'/Overlay': ['myTestMachineVeryRare'],
                     '/Overlay/myTestMachineVeryRare': ['other_energy_tev', '981tev'],
                     '/Overlay/myTestMachineVeryRare/981tev': ['testDetMod']}
    ops_mock = Mock()
    ops_mock.getSections.side_effect = lambda path: S_OK(section_dict[path])
    self.olin._ops = ops_mock
    self.olin.machine = 'myTestMachineVeryRare'
    with patch.object(inspect.getmodule(OverlayInput), 'allowedBkg', new=Mock(return_value=S_OK(-147))):
      assertDiracFailsWith(self.olin._checkFinalConsistency(), 'no proper production id found', self)

  def test_checkfinalconsistency(self):
    self.olin.detectorModel = 'testDetMod'
    self.olin.energy = 981000
    section_dict = {'/Overlay': ['myTestMachineVeryRare'],
                     '/Overlay/myTestMachineVeryRare': ['other_energy_tev', '981tev'],
                     '/Overlay/myTestMachineVeryRare/981tev': ['testDetMod']}
    ops_mock = Mock()
    ops_mock.getSections.side_effect = lambda path: S_OK(section_dict[path])
    self.olin._ops = ops_mock
    self.olin.machine = 'myTestMachineVeryRare'
    with patch.object(inspect.getmodule(OverlayInput), 'allowedBkg', new=Mock(return_value=S_OK(13987))):
      assertDiracSucceeds(self.olin._checkFinalConsistency(), self)

  def test_printAvailableOptions(self):
    """Tests for printAvailableOptions.

    As there are no return values we check that the proper log messages are printed to error or notice.
    """
    confMock = Mock()
    confMock.getConfigurationTree.return_value = S_ERROR('')
    logMock = Mock(name='Logger')
    with patch.object(inspect.getmodule(OverlayInput), 'gConfig', new=confMock), \
         patch.object(inspect.getmodule(OverlayInput), 'LOG', new=logMock):
      self.olin.printAvailableOptions()
    logMock.error.assert_called_once()

    confMock = Mock()
    confMock.getConfigurationTree.return_value = S_OK({'/Operations/Defaults/Overlay/CLIC/1tev/DET/SUSY': 12,
                                                       '/Operations/Defaults/Overlay/Not_CLIC/1tev/DET/BKG_TYPE': 14,
                                                       '/Operations/Defaults/Overlay/CLIC/1tev/Other_DET/BKG_TYPE': 15,
                                                       '/Operations/Defaults/Overlay/CLIC/1tev/DET': 13,
                                                     })

    logMock.reset()
    with patch.object(inspect.getmodule(OverlayInput), 'gConfig', new=confMock), \
         patch.object(inspect.getmodule(OverlayInput), 'LOG', new=logMock):
      self.olin.printAvailableOptions()
    logMock.info.assert_not_called()  # make sure we use notice

    logMock.reset()
    with patch.object(inspect.getmodule(OverlayInput), 'gConfig', new=confMock), \
         patch.object(inspect.getmodule(OverlayInput), 'LOG', new=logMock):
      self.olin.printAvailableOptions(detModel='DET')
    logMock.info.assert_not_called()
    logMock.notice.assert_any_call('Printing options compatible with')
    logMock.notice.assert_any_call(' * DetModel = DET')

    logMock.reset()
    with patch.object(inspect.getmodule(OverlayInput), 'gConfig', new=confMock), \
         patch.object(inspect.getmodule(OverlayInput), 'LOG', new=logMock):
      self.olin.printAvailableOptions(energy=2000)
    logMock.info.assert_not_called()
    logMock.notice.assert_any_call('Printing options compatible with')
    logMock.notice.assert_any_call(' * Energy = 2tev')

    logMock.reset()
    with patch.object(inspect.getmodule(OverlayInput), 'gConfig', new=confMock), \
         patch.object(inspect.getmodule(OverlayInput), 'LOG', new=logMock):
      self.olin.printAvailableOptions(machine='CLIC')
    logMock.info.assert_not_called()
    logMock.notice.assert_any_call('Printing options compatible with')
    logMock.notice.assert_any_call(' * Machine = CLIC')

    logMock.reset()
    with patch.object(inspect.getmodule(OverlayInput), 'gConfig', new=confMock), \
         patch.object(inspect.getmodule(OverlayInput), 'LOG', new=logMock):
      self.olin.printAvailableOptions(energy=2000, machine='FCCee', detModel='NO_DET')
    logMock.notice.assert_any_call('No overlay options compatible with your selection')



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Tests for .Core.Utilities.ProductionData.constructProductionLFNs."""

from __future__ import print_function
from __future__ import absolute_import
import unittest
from ILCDIRAC.Core.Utilities.ProductionData import constructProductionLFNs

__RCSID__ = "$Id$"


class ProductionOutputDataTestCase(unittest.TestCase):
  """Base class for the test cases."""

  def setUp(self):
    pass

  def tearDown(self):
    pass

  def test_contructProductionLFNBad(self):
    """test ProductionOutputData construct Bad LFN.................................................."""
    commons = {}
    result = constructProductionLFNs(commons)
    self.assertEqual(result['OK'], False)

  def test_contructProductionLFNstdhep(self):
    """test ProductionOutputData construct stdhep LFN..............................................."""
    commons = {}
    commons['PRODUCTION_ID'] = 12345
    commons['JOB_ID'] = 1234
    commons['outputList'] = [{'outputFile': "something_gen.stdhep",
                              'outputPath': '/ilc/prod/clic/test/gen'}]
    result = constructProductionLFNs(commons)

    self.assertEqual(result['OK'], True)

  def test_contructProductionLFNsim(self):
    """test ProductionOutputData construct sim LFN.................................................."""
    commons = {}
    commons['PRODUCTION_ID'] = 12345
    commons['JOB_ID'] = 1234
    commons['outputList'] = [{'outputFile': "something_sim.slcio",
                              'outputPath': '/ilc/prod/clic/test/SIM'}]
    result = constructProductionLFNs(commons)
    self.assertEqual(result['OK'], True)

  def test_contructProductionLFNrec(self):
    """test ProductionOutputData construct rec LFN.................................................."""
    commons = {}
    commons['PRODUCTION_ID'] = 12345
    commons['JOB_ID'] = 1234
    commons['outputList'] = [{'outputFile': "something_rec.slcio",
                              'outputPath': '/ilc/prod/clic/test/REC'},
                             {'outputFile': "something_dst.slcio",
                              'outputPath': '/ilc/prod/clic/test/DST'}]
    result = constructProductionLFNs(commons)
    self.assertEqual(result['OK'], True)

  def test_contructProductionLFNoutput(self):
    """test ProductionOutputData construct out LFN.................................................."""
    commons = {}
    commons['PRODUCTION_ID'] = 12345
    commons['JOB_ID'] = 1234
    commons['outputList'] = [{'outputFile': "something_gen.stdhep",
                              'outputPath': '/ilc/prod/clic/test/gen'}]
    result = constructProductionLFNs(commons)
    res = {'ProductionOutputData': ["/ilc/prod/clic/test/gen/00012345/001/something_gen_12345_1234.stdhep"],
           'LogFilePath': ["/ilc/prod/clic/test/gen/00012345/LOG/001"],
           'LogTargetPath': ["/ilc/prod/clic/test/gen/LOG/00012345/00012345_1234.tar"]}
    for key in res.keys():
      self.assertEqual(result['Value'][key], res[key])


def runTests():
  """runs all the tests."""
  suite = unittest.defaultTestLoader.loadTestsFromTestCase(ProductionOutputDataTestCase)
  testResult = unittest.TextTestRunner(verbosity=2).run(suite)
  print(testResult)


if __name__ == '__main__':
  runTests()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""
LCSIM: Reconstruction after SLIC Simulation
"""

from __future__ import absolute_import
import os
import types

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Workflow.Parameter import Parameter

from ILCDIRAC.Interfaces.API.NewInterface.LCApplication import LCApplication
from ILCDIRAC.Core.Utilities.CheckXMLValidity import checkXMLValidity
from ILCDIRAC.Core.Utilities.InstalledFiles import Exists
import six

LOG = gLogger.getSubLogger(__name__)
__RCSID__ = "$Id$"


class LCSIM(LCApplication):
  """Call LCSIM Reconstructor (after SLIC Simulation)

  Usage:

  >>> slic = SLIC()
  >>> lcsim = LCSIM()
  >>> lcsim.getInputFromApp(slic)
  >>> lcsim.setSteeringFile("MySteeringFile.xml")
  >>> lcsim.setStartFrom(10)

  Use :func:`setExtraCLIArguments` to add command line arguments to the lcsim call
  """

  def __init__(self, paramdict=None):

    self.extraParams = ''
    self.aliasProperties = ''
    self.trackingStrategy = ''
    self.detectorModel = ''
    super(LCSIM, self).__init__(paramdict)
    # Those 5 need to come after default constructor
    self._modulename = 'LCSIMAnalysis'
    self._moduledescription = 'Module to run LCSIM'
    self.appname = 'lcsim'
    self.datatype = 'REC'
    self.detectortype = 'SID'

  def setOutputRecFile(self, outputRecFile, path=None):
    """Optional: Define output rec file for LCSIM Used only in production
    context. Use :func:`UserJob.setOutputData
    <ILCDIRAC.Interfaces.API.NewInterface.UserJob.UserJob.setOutputData>` if you
    want to keep the file on the grid.

    :param str outputRecFile: output rec file for LCSIM
    :param str path: Path where to store the file.

    """
    self._checkArgs({'outputRecFile': (str,)})
    self.outputRecFile = outputRecFile
    self.prodparameters[self.outputRecFile] = {}
    self.prodparameters[self.outputRecFile]['datatype'] = 'REC'
    if path:
      self.outputRecPath = path

  def setOutputDstFile(self, outputDstFile, path=None):
    """Optional: Define output dst file for LCSIM.Used only in production
    context. Use :func:`UserJob.setOutputData
    <ILCDIRAC.Interfaces.API.NewInterface.UserJob.UserJob.setOutputData>` if you
    want to keep the file on the grid.

    :param str outputDstFile: output dst file for LCSIM
    :param str path: Path where to store the file.

    """
    self._checkArgs({'outputDstFile': (str,)})
    self.outputDstFile = outputDstFile
    self.prodparameters[self.outputDstFile] = {}
    self.prodparameters[self.outputDstFile]['datatype'] = 'DST'
    if path:
      self.outputDstPath = path

  def setAliasProperties(self, alias):
    """ Optional: Define the path to the alias.properties file name that will be used

    :param str alias: Path to the alias.properties file name that will be used
    """
    self._checkArgs({'alias': (str,)})

    self.aliasProperties = alias
    if os.path.exists(alias) or alias.lower().count("lfn:"):
      self.inputSB.append(alias)

  def setDetectorModel(self, model):
    """Detector Model to use.

    :param str model: name, zip file, or lfn that points to the detector model
    """
    self._checkArgs({'model': (str,)})
    self.detectorModel = model
    if os.path.exists(model) or model.lower().count("lfn:"):
      self.inputSB.append(model)

  def setTrackingStrategy(self, trackingstrategy):
    """ Optional: Define the tracking strategy to use.

    :param str trackingstrategy: path to the trackingstrategy file to use. If not called, will use whatever is
                                    in the steering file

    """
    self._checkArgs({'trackingstrategy': (str,)})
    self.trackingStrategy = trackingstrategy
    if os.path.exists(self.trackingStrategy) or self.trackingStrategy.lower().count('lfn:'):
      self.inputSB.append(self.trackingStrategy)

  def setExtraParams(self, extraparams):
    """ Optional: Define command line parameters to pass to java

    :param str extraparams: Command line parameters to pass to java
    """
    self._checkArgs({'extraparams': (str,)})

    self.extraParams = extraparams

  def willRunSLICPandora(self):
    """You need this if you plan on running :mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.SLICPandora`"""
    self.willBeCut = True

  def _userjobmodules(self, stepdefinition):
    res1 = self._setApplicationModuleAndParameters(stepdefinition)
    res2 = self._setUserJobFinalization(stepdefinition)
    if not res1["OK"] or not res2["OK"]:
      return S_ERROR('userjobmodules failed')
    return S_OK()

  def _prodjobmodules(self, stepdefinition):
    res1 = self._setApplicationModuleAndParameters(stepdefinition)
    res2 = self._setOutputComputeDataList(stepdefinition)
    if not res1["OK"] or not res2["OK"]:
      return S_ERROR('prodjobmodules failed')
    return S_OK()

  def _checkConsistency(self, job=None):

    if not self.energy:
      LOG.info('Energy set to 0 !')

    if not self.numberOfEvents:
      LOG.info('Number of events set to 0 !')

    if not self.version:
      return S_ERROR('No version found')

    if self.steeringFile:
      if not os.path.exists(self.steeringFile) and not self.steeringFile.lower().count("lfn:"):
        res = Exists(self.steeringFile)
        if not res['OK']:
          return res
      if os.path.exists(self.steeringFile):
        res = checkXMLValidity(self.steeringFile)
        if not res['OK']:
          return S_ERROR("Supplied steering file cannot be read by XML parser: %s" % (res['Message']))
    if self.trackingStrategy:
      if not os.path.exists(self.trackingStrategy) and not self.trackingStrategy.lower().count("lfn:"):
        res = Exists(self.trackingStrategy)
        if not res['OK']:
          return res

    if self.detectorModel:
      if not self.detectorModel.lower().count(".zip"):
        return S_ERROR("setDetectorModel: You HAVE to pass an existing .zip file, either as local file or as LFN. \
        Or use the alias.properties.")

    #res = self._checkRequiredApp()
    # if not res['OK']:
    #  return res

    if self._jobtype != 'User':
      #slicp = False
      if self._inputapp and not self.outputFile and not self.willBeCut:
        for app in self._inputapp:
          if app.appname in ['slicpandora', 'marlin']:
            self._listofoutput.append({"outputFile": "@{outputREC}", "outputPath": "@{outputPathREC}",
                                       "outputDataSE": '@{OutputSE}'})
            self._listofoutput.append({"outputFile": "@{outputDST}", "outputPath": "@{outputPathDST}",
                                       "outputDataSE": '@{OutputSE}'})
            #slicp = True
            break
      self.prodparameters['detectorType'] = self.detectortype
      self.prodparameters['lcsim_steeringfile'] = self.steeringFile
      self.prodparameters['lcsim_trackingstrategy'] = self.trackingStrategy

      # if not slicp:
      #  self._listofoutput.append({"outputFile":"@{OutputFile}","outputPath":"@{OutputPath}","outputDataSE":'@{OutputSE}'})

    return S_OK()

  def _applicationModule(self):

    md1 = self._createModuleDefinition()
    md1.addParameter(Parameter("extraparams", "", "string", "", "", False, False,
                               "Command line parameters to pass to java"))
    md1.addParameter(Parameter("aliasproperties", "", "string", "", "", False, False,
                               "Path to the alias.properties file name that will be used"))
    md1.addParameter(Parameter("debug", False, "bool", "", "", False, False,
                               "debug mode"))
    md1.addParameter(Parameter("detectorModel", "", "string", "", "", False, False,
                               "detector model zip file"))
    md1.addParameter(Parameter("trackingstrategy", "", "string", "", "", False, False,
                               "trackingstrategy"))
    return md1

  def _applicationModuleValues(self, moduleinstance):

    moduleinstance.setValue("extraparams", self.extraParams)
    moduleinstance.setValue("aliasproperties", self.aliasProperties)
    moduleinstance.setValue("debug", self.debug)
    moduleinstance.setValue("detectorModel", self.detectorModel)
    moduleinstance.setValue("trackingstrategy", self.trackingStrategy)

  def _checkWorkflowConsistency(self):
    return self._checkRequiredApp()

  def _resolveLinkedStepParameters(self, stepinstance):
    if isinstance(self._linkedidx, six.integer_types):
      self._inputappstep = self._jobsteps[self._linkedidx]
    if self._inputappstep:
      stepinstance.setLink("InputFile", self._inputappstep.getType(), "OutputFile")
    return S_OK()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""This is the worst piece of code ever NEEDED: just run some CPU intensive code to waste CPU time.

:author: Stephane Poss
:since: Jul 26, 2011
"""

from __future__ import absolute_import
from datetime import datetime
from math import log
from DIRAC import S_OK, S_ERROR

__RCSID__ = "$Id$"


def wasteCPUCycles(timecut):
  """Waste, waste, and waste more CPU."""
  number = 1e31
  first = datetime.utcnow()
  try:
    while (datetime.utcnow() - first).total_seconds() < timecut:
      try:
        number = log(number)
      except ValueError as x:
        return S_ERROR("Failed to waste %s CPU seconds:%s" % (timecut, str(x)))
      if number <= 0:
        number = -number + 4
  except Exception as e:  # pylint: disable=broad-except
    return S_ERROR("Failed to waste %s CPU seconds, OtherException: %r" % (timecut, e))
  return S_OK("Successfully wasted %s seconds" % timecut)



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""
OverlayInput : Helper call to define Overlay processor/driver inputs
"""
# pylint: disable=expression-not-assigned

from __future__ import absolute_import
import types
from collections import defaultdict

from ILCDIRAC.Interfaces.API.NewInterface.LCUtilityApplication import LCUtilityApplication
from ILCDIRAC.Workflow.Modules.OverlayInput import allowedBkg
from ILCDIRAC.Core.Utilities.OverlayFiles import energyWithLowerCaseUnit
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient
from DIRAC.Core.Workflow.Parameter import Parameter
from DIRAC import S_OK, S_ERROR, gLogger, gConfig

LOG = gLogger.getSubLogger(__name__)
__RCSID__ = "$Id$"


class OverlayInput(LCUtilityApplication):
  """Helper call to define Overlay processor/driver inputs.

  Example:

  >>> over = OverlayInput()
  >>> over.setBXOverlay(300)
  >>> over.setGGToHadInt(3.2)
  >>> over.setNumberOfSignalEventsPerJob(10)
  >>> over.setBackgroundType("gghad")
  >>> over.setMachine("clic_opt")
  >>> over.setDetectorModel("CLICdet_o3_v14")
  >>> over.setEnergy("3000")

  See list of available options:
  >>> over = OverlayInput()
  >>> over.printAvailableOptions()
  """

  def __init__(self, paramdict=None):
    self._ops = Operations()
    self.bxToOverlay = None
    self.numberOfGGToHadronInteractions = 0
    self.numberOfSignalEventsPerJob = 0
    self.backgroundEventType = ''
    self.prodID = 0
    self.machine = 'clic_cdr'
    self.detectorModel = ''
    self.useEnergyForFileLookup = True
    super(OverlayInput, self).__init__(paramdict)
    self.version = '1'
    self._modulename = "OverlayInput"
    self.appname = self._modulename
    self._moduledescription = 'Helper call to define Overlay processor/driver inputs'
    self.accountInProduction = False
    self._paramsToExclude.append('_ops')
    self.pathToOverlayFiles = ''
    self.processorName = ''

  def setMachine(self, machine):
    """Define the machine to use."""
    self._checkArgs({'machine': (str,)})
    self.machine = machine

  def setProdID(self, pid):
    """Define the prodID to use as input, experts only."""
    self._checkArgs({'pid': int})
    self.prodID = pid
    return S_OK()

  def setUseEnergyForFileLookup(self, useEnergyForFileLookup):
    """Set the flag to use the energy meta data in the search of the background files.

    Disable the energy when you want to use files created for a different energy than the signal events

    :param bool useEnergyForFileLookup: Use the Energy in the metadata search or not
    """
    self._checkArgs({'useEnergyForFileLookup': bool})
    self.useEnergyForFileLookup = useEnergyForFileLookup
    return S_OK()

  def setOverlayBXPerSigEvt(self, bxoverlay):
    """Define number bunch crossings to overlay for each signal event.

    This is used to determine the number of required overlay events.
    It does not modify any of the actual application parameters using the overly input.
    Alias for :func:`setBXOverlay`

    :param int bxoverlay: Bunch crossings to overlay.
    """
    self._checkArgs({'bxoverlay': int})
    self.bxToOverlay = bxoverlay
    return S_OK()

  def setBXOverlay(self, bxoverlay):
    """Define number bunch crossings to overlay for each signal event.

    This is used to determine the number of required overlay events.
    It does not modify any of the actual application parameters using the overly input.

    :param int bxoverlay: Bunch crossings to overlay.
    """
    return self.setOverlayBXPerSigEvt(bxoverlay)

  def setOverlayEvtsPerBX(self, ggtohadint):
    """Define the number of overlay events per bunch crossing.

    This is used to determine the number of required overlay events.
    It does not modify any of the actual application parameters using the overly input.

    :param float ggtohadint: optional number of overlay events interactions per bunch crossing
    """
    self._checkArgs({'ggtohadint': float})
    self.numberOfGGToHadronInteractions = ggtohadint
    return S_OK()

  def setGGToHadInt(self, ggtohadint):
    """Define the number of overlay events per bunch crossing.

    This is used to determine the number of required overlay events.
    It does not modify any of the actual application parameters using the overly input.

    Alias for :func:`setOverlayEvtsPerBX`

    :param float ggtohadint: optional number of overlay events interactions per bunch crossing
    """
    return self.setOverlayEvtsPerBX(ggtohadint)

  def setNbSigEvtsPerJob(self, nbsigevtsperjob):
    """Set the number of signal events per job.

    This is used to determine the number of required overlay events.
    It does not modify any of the actual application parameters using the overly input.

    :param int nbsigevtsperjob: Number of signal events per job
    """
    self._checkArgs({'nbsigevtsperjob': int})

    self.numberOfSignalEventsPerJob = nbsigevtsperjob
    return S_OK()

  def setDetectorModel(self, detectormodel):
    """Set the detector type for the background files.

    Files are defined in the ConfigurationSystem: Operations/Overlay/<Accelerator>/<energy>/<Detector>

    :param str detectormodel: Detector type
    """
    self._checkArgs({'detectormodel': (str,)})

    self.detectorModel = detectormodel
    return S_OK()

  def setPathToFiles(self, path):
    """Set the path to where the overlay files are located.

    Setting this option will ignore all other settings!

    :param str path: LFN path to the folder containing the overlay files
    """
    self._checkArgs({'path': (str,)})
    self.pathToOverlayFiles = path
    return S_OK()

  def setBkgEvtType(self, backgroundEventType):
    """Define the background type.

    .. deprecated:: 23r0
       Use :func:`setBackgroundType` instead

    :param str backgroundEventType: Background type.
    """
    self._checkArgs({'backgroundEventType': (str,)})

    self.backgroundEventType = backgroundEventType
    return S_OK()

  def setBackgroundType(self, backgroundType):
    """Define the background type.

    :param str backgroundType: Background type.
    """
    return self.setBkgEvtType(backgroundType)

  def setProcessorName(self, processorName):
    """Set the processorName to set the input files for.

    Necessary if multiple invocations of the overlay processor happen in marlin for example.
    Different processors must use different background types

    :param str processorName: Name of the Processor these input files are for
    """
    self._checkArgs({'processorName': (str,)})
    self.processorName = processorName
    return S_OK()

  def setNumberOfSignalEventsPerJob(self, numberSignalEvents):
    """Alternative to :func:`setNbSigEvtsPerJob`

    Number used to determine the number of background files needed.

    :param int numberSignalEvents: Number of signal events per job
    """
    return self.setNbSigEvtsPerJob(numberSignalEvents)

  def _applicationModule(self):
    m1 = self._createModuleDefinition()
    m1.addParameter(Parameter("BXOverlay", 0, "float", "", "", False, False,
                              "Bunch crossings to overlay"))
    m1.addParameter(
        Parameter(
            "ggtohadint",
            0,
            "float",
            "",
            "",
            False,
            False,
            "Optional number of gamma gamma -> hadrons interactions per bunch crossing, default is 3.2"))
    m1.addParameter(Parameter("NbSigEvtsPerJob", 0, "int", "", "", False, False,
                              "Number of signal events per job"))
    m1.addParameter(Parameter("prodid", 0, "int", "", "", False, False,
                              "ProdID to use"))
    m1.addParameter(Parameter("BkgEvtType", "", "string", "", "", False, False,
                              "Background type."))
    m1.addParameter(Parameter("detectormodel", "", "string", "", "", False, False,
                              "Detector type."))
    m1.addParameter(Parameter("machine", "", "string", "", "", False, False,
                              "machine: clic_cdr or ilc_dbd"))
    m1.addParameter(Parameter("useEnergyForFileLookup", True, "bool", "", "", False, False,
                              "useEnergy to look for background files: True or False"))
    m1.addParameter(Parameter("pathToOverlayFiles", "", "string", "", "", False, False,
                              "use overlay files from this path"))
    m1.addParameter(Parameter("processorName", "", "string", "", "", False, False,
                              "Processor Name"))

    m1.addParameter(Parameter("debug", False, "bool", "", "", False, False, "debug mode"))
    return m1

  def _applicationModuleValues(self, moduleinstance):
    moduleinstance.setValue("BXOverlay", self.bxToOverlay)
    moduleinstance.setValue('ggtohadint', self.numberOfGGToHadronInteractions)
    moduleinstance.setValue('NbSigEvtsPerJob', self.numberOfSignalEventsPerJob)
    moduleinstance.setValue('prodid', self.prodID)
    moduleinstance.setValue('BkgEvtType', self.backgroundEventType)
    moduleinstance.setValue('detectormodel', self.detectorModel)
    moduleinstance.setValue('debug', self.debug)
    moduleinstance.setValue('machine', self.machine)
    moduleinstance.setValue('useEnergyForFileLookup', self.useEnergyForFileLookup)
    moduleinstance.setValue('pathToOverlayFiles', self.pathToOverlayFiles)
    moduleinstance.setValue('processorName', self.processorName)

  def _userjobmodules(self, stepdefinition):
    res1 = self._setApplicationModuleAndParameters(stepdefinition)
    if not res1["OK"]:
      return S_ERROR('userjobmodules failed')
    return S_OK()

  def _prodjobmodules(self, stepdefinition):
    res1 = self._setApplicationModuleAndParameters(stepdefinition)
    if not res1["OK"]:
      return S_ERROR('prodjobmodules failed')
    return S_OK()

  def _addParametersToStep(self, stepdefinition):
    res = self._addBaseParameters(stepdefinition)
    if not res["OK"]:
      return S_ERROR("Failed to set base parameters")
    return S_OK()

  def _checkConsistency(self, job=None):
    """Checks that all needed parameters are set."""
    if self.pathToOverlayFiles:
      res = FileCatalogClient().findFilesByMetadata({}, self.pathToOverlayFiles)
      if not res['OK']:
        return res
      LOG.notice("Found %i files in path %s" % (len(res['Value']), self.pathToOverlayFiles))
      if not res['Value']:
        return S_ERROR("OverlayInput: PathToFiles is specified, but there are no files in that path")

    if not self.bxToOverlay:
      return S_ERROR("Number of overlay bunch crossings not defined")

    if not self.numberOfGGToHadronInteractions:
      return S_ERROR("Number of background events per bunch crossing is not defined")

    if not self.backgroundEventType:
      return S_ERROR("Background event type is not defined: Chose one gghad, aa_lowpt, ...")

    if self._jobtype == 'User':
      if not self.numberOfSignalEventsPerJob:
        return S_ERROR("Number of signal event per job is not defined")
    else:
      self.prodparameters['detectorModel'] = self.detectorModel
      self.prodparameters['BXOverlay'] = self.bxToOverlay
      self.prodparameters['GGtoHadInt'] = self.numberOfGGToHadronInteractions

    return S_OK()

  @staticmethod
  def printAvailableOptions(machine=None, energy=None, detModel=None):
    """Print a list of the available options for each machine, energy and detector model.

    The parameters can be used to filter the output

    :param str machine: only list options for this machine
    :param str detModel: only list options for this detector model
    :param energy: only list options for this energy
    :type energy: int or float
    """
    overTree = gConfig.getConfigurationTree('/Operations/Defaults/Overlay')
    if not overTree['OK']:
      LOG.error('Could not find the configuration section "/Operations/Defaults/Overlay"')
      return

    if energy:
      energy = energyWithLowerCaseUnit(energy)

    # Dictionary of machines, energy, detModel, backgroundType
    availableOptions = defaultdict(lambda: defaultdict(lambda: defaultdict(set)))
    for value in overTree['Value']:
      values = value.strip('/').split('/')
      if len(values) <= 6:
        continue
      theMachine = values[3]
      theEnergy = values[4]
      theDetModel = values[5]
      backgroundType = values[6]
      if machine and machine != theMachine:
        continue
      if energy and energy != theEnergy:
        continue
      if detModel and detModel != theDetModel:
        continue
      availableOptions[theMachine][theEnergy][theDetModel].add(backgroundType)

    if machine or energy or detModel:
      if availableOptions:
        LOG.notice('Printing options compatible with')
      else:
        LOG.notice('No overlay options compatible with your selection')
      LOG.notice(' * Machine = %s' % machine) if machine else False
      LOG.notice(' * Energy = %s' % energy) if energy else False
      LOG.notice(' * DetModel = %s' % detModel) if detModel else False
    else:
      LOG.notice('All available overlay combinations')
    for theMachine, energies in availableOptions.items():
      LOG.notice('Machine: %s' % theMachine)
      for theEnergy, detModels in energies.items():
        LOG.notice('    %s' % theEnergy)
        for theDetModel, backgrounds in detModels.items():
          LOG.notice('        %s: %s' % (theDetModel, ', '.join(backgrounds)))

  def _checkFinalConsistency(self):
    """Check consistency before submission.

    The overlay files for the specifed energy must exist. Print all available overlay options on error
    """
    res = self.__checkFinalConsistency()
    if res['OK']:
      return res

    self.printAvailableOptions(machine=self.machine)
    self.printAvailableOptions(energy=self.energy)
    self.printAvailableOptions(detModel=self.detectorModel)
    return res

  def __checkFinalConsistency(self):
    """Check consistency of overlay options."""
    if self.pathToOverlayFiles:
      return S_OK()  # can ignore other parameter

    if not self.energy:
      return S_ERROR("Energy MUST be specified for the overlay")

    res = self._ops.getSections('/Overlay')
    if not res['OK']:
      return S_ERROR("Could not resolve the CS path to the overlay specifications")
    sections = res['Value']
    if self.machine not in sections:
      return S_ERROR('Machine %r does not have overlay data' % self.machine)

    energytouse = energyWithLowerCaseUnit(self.energy)
    res = self._ops.getSections("/Overlay/%s" % self.machine)
    if energytouse not in res['Value']:
      return S_ERROR("No overlay files corresponding to %s" % energytouse)

    res = self._ops.getSections("/Overlay/%s/%s" % (self.machine, energytouse))
    if not res['OK']:
      return S_ERROR("Could not find the detector models")

    if self.detectorModel not in res['Value']:
      return S_ERROR('Detector model %r has no overlay data with energy %r and %r' %
                     (self.detectorModel, self.energy, self.machine))

    res = allowedBkg(self.backgroundEventType, energytouse, detectormodel=self.detectorModel, machine=self.machine)
    if not res['OK']:
      return res
    if res['Value'] < 0:
      return S_ERROR("No proper production ID found")
    return S_OK()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Execute the FileStatusTransformationAgent checks on a given Transformation ID.

Check for consistency in the status of replication of moving transformation.

Example::

  dirac-ilc-filestatus-transformation <transformationID> -x

Options:
   -x, --enable         Perform actions on tasks, requests and files
"""

from __future__ import print_function
from __future__ import absolute_import
__RCSID__ = "$Id$"

from DIRAC.Core.Base import Script
from DIRAC.Core.Utilities.DIRACScript import DIRACScript
from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC import exit as dexit


class _Params(object):
  """parameters object."""

  def __init__(self):
    self.transID = None
    self.enabled = False

  def setTransID(self, transID):
    self.transID = transID

  def setEnabled(self, opt):
    self.enabled = True
    return S_OK()

  def registerSwitches(self):
    Script.registerSwitch("x", "enable", "perform delete operations on file catalog", self.setEnabled)
    Script.setUsageMessage("""%s <transformationID> -x""" % Script.scriptName)

  def checkSettings(self):
    """parse arguments."""

    args = Script.getPositionalArgs()
    if len(args) < 1:
      return S_ERROR()

    self.setTransID(args[0])

    return S_OK()


@DIRACScript()
def main():
  """read commands line params and run FST agent for a given transformation ID."""
  params = _Params()
  params.registerSwitches()
  Script.parseCommandLine()
  if not params.checkSettings()['OK']:
    Script.showHelp()
    dexit(1)

  from ILCDIRAC.ILCTransformationSystem.Agent.FileStatusTransformationAgent import FileStatusTransformationAgent
  fstAgent = FileStatusTransformationAgent('ILCTransformation/FileStatusTransformationAgent',
                                           'ILCTransformation/FileStatusTransformationAgent',
                                           'dirac-ilc-filestatus-transformation')
  fstAgent.log = gLogger
  fstAgent.enabled = params.enabled

  res = fstAgent.getTransformations(transID=params.transID)
  if not res['OK']:
    dexit(1)

  if not res['Value']:
    print("Transformation Not Found")
    dexit(1)

  trans = res['Value'][0]

  res = fstAgent.processTransformation(
      int(params.transID), trans['SourceSE'], trans['TargetSE'], trans['DataTransType'])
  if not res["OK"]:
    dexit(1)

  dexit(0)


if __name__ == "__main__":
  main()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""For any input file, try to determine from the FC the number of events / luminosity / event type.

:author: S. Poss
:since: Nov 2, 2010
"""

from __future__ import absolute_import
__RCSID__ = "$Id$"

from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient
import os

from DIRAC import gLogger, S_OK, S_ERROR

LOG = gLogger.getSubLogger(__name__)


def getNumberOfEvents(inputfile):
  """Find from the FileCatalog the number of events in a file.

  :param list inputfile: list of file LFNs
  :returns: S_OK() with dictionary containing information on the number of events with the keys:

     * nbevts
     * lumi
     * EvtType
  """

  files = inputfile
  flist = {}
  for myfile in files:
    if not myfile:
      continue
    bpath = os.path.dirname(myfile)
    if bpath not in flist:
      flist[bpath] = [myfile]
    else:
      flist[bpath].append(myfile)

  fc = FileCatalogClient()
  nbevts = {}
  luminosity = 0
  numberofevents = 0
  evttype = ''
  others = {}
  completeFailure = True

  for path, files in flist.items():
    found_nbevts = False
    found_lumi = False

    if len(files) == 1:
      res = fc.getFileUserMetadata(files[0])
      if not res['OK']:
        LOG.warn("Failed to get Metadata from file: %s, because: %s" % (files[0], res['Message']))
      else:
        tags = res['Value']
        if "NumberOfEvents" in tags and not found_nbevts:
          numberofevents += int(tags["NumberOfEvents"])
          found_nbevts = True
          completeFailure = False
        if "Luminosity" in tags and not found_lumi:
          luminosity += float(tags["Luminosity"])
          found_lumi = True
        others.update(tags)
        if found_nbevts:
          continue

    res = fc.getDirectoryUserMetadata(path)
    if res['OK']:
      tags = res['Value']
      if "NumberOfEvents" in tags and not found_nbevts:
        numberofevents += len(files) * int(tags["NumberOfEvents"])
        found_nbevts = True
        completeFailure = False
      if "Luminosity" in tags and not found_lumi:
        luminosity += len(files) * float(tags["Luminosity"])
        found_lumi = True

      evttype = tags.get("EvtType", evttype)
      others.update(tags)
      if found_nbevts:
        continue
    else:
      LOG.warn("Failed to get Metadata from path: %s, because: %s" % (path, res['Message']))

    for myfile in files:
      res = fc.getFileUserMetadata(myfile)
      if not res['OK']:
        LOG.warn("Failed to get Metadata from file: %s, because: %s" % (myfile, res['Message']))
        continue
      tags = res['Value']
      if "NumberOfEvents" in tags:
        numberofevents += int(tags["NumberOfEvents"])
        completeFailure = False
      if "Luminosity" in tags and not found_lumi:
        luminosity += float(tags["Luminosity"])
      others.update(tags)

  nbevts['nbevts'] = numberofevents
  nbevts['lumi'] = luminosity
  nbevts['EvtType'] = evttype
  if 'NumberOfEvents' in others:
    del others['NumberOfEvents']
  if 'Luminosity' in others:
    del others['Luminosity']
  nbevts['AdditionalMeta'] = others

  if completeFailure:
    LOG.warn("Did not obtain NumberOfEvents from FileCatalog")
    return S_ERROR("Failed to get Number of Events")

  return S_OK(nbevts)



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Interface to the processlist.whiz that contains all the processes known to WHIZARD.

:author: S. Poss
:since: Sep 21, 2010
"""

from __future__ import absolute_import
__RCSID__ = "$Id$"

from DIRAC import S_OK, S_ERROR, gLogger
from diraccfg import CFG
from pprint import pprint
import os
import tempfile
import shutil
import subprocess

LOG = gLogger.getSubLogger(__name__)


class ProcessList(object):
  """The ProcessList uses internally the CFG utility to store the processes and their properties."""

  def __init__(self, location):
    self.cfg = CFG()
    self.location = location
    self.goodProcessList = True
    if os.path.exists(self.location):
      self.cfg.loadFromFile(self.location)
      if not self.cfg.existsKey('Processes'):
        self.cfg.createNewSection('Processes')
    else:
      self.goodProcessList = False

  def _writeProcessList(self, path):
    """Write to text."""
    handle, tmpName = tempfile.mkstemp()
    written = self.cfg.writeToFile(tmpName)
    os.close(handle)
    if not written:
      if os.path.exists(tmpName):
        os.remove(tmpName)
      return written
    if os.path.exists(path):
      LOG.debug("Replacing %s" % path)
    try:
      shutil.move(tmpName, path)
      return True
    except OSError as err:
      LOG.error("Failed to overwrite process list.", err)
      LOG.info("If your process list is corrupted a backup can be found %s" % tmpName)
      return False

  def isOK(self):
    """Check if the content is OK."""
    return self.goodProcessList

  def updateProcessList(self, processes):
    """Adds a new entry or updates an existing one.

    :param dict processes: dictionary of processes to treat
    """
    LOG.verbose("Updating process list:")
    for process, mydict in processes.items():
      if not self._existsProcess(process):
        self._addEntry(process, mydict)
        # return res
      else:
        LOG.warn("Process %s already defined in ProcessList, will replace it" % process)
        self.cfg.deleteKey("Processes/%s" % process)
        self._addEntry(process, mydict)
        # return res
    LOG.verbose("Done Updating process list")
    return S_OK()

  def _addEntry(self, process, processdic):
    """Adds a new entry."""
    if not self.cfg.isSection("Processes/%s" % process):
      self.cfg.createNewSection("Processes/%s" % process)
    self.cfg.setOption("Processes/%s/TarBallCSPath" % process, processdic['TarBallCSPath'])
    self.cfg.setOption("Processes/%s/Detail" % process, processdic['Detail'])
    self.cfg.setOption("Processes/%s/Generator" % process, processdic['Generator'])
    self.cfg.setOption("Processes/%s/Model" % process, processdic['Model'])
    self.cfg.setOption("Processes/%s/Restrictions" % process, processdic['Restrictions'])
    self.cfg.setOption("Processes/%s/InFile" % process, processdic['InFile'])
    cross_section = 0
    if 'CrossSection' in processdic:
      cross_section = processdic["CrossSection"]
    self.cfg.setOption("Processes/%s/CrossSection" % process, cross_section)
    return S_OK()

  def getCSPath(self, process):
    """Return the path to the TarBall (for install)

    :param str process: process to look for
    """
    return self.cfg.getOption("Processes/%s/TarBallCSPath" % process, None)

  def getInFile(self, process):
    """Get the associated whizard.in file to the process."""
    return self.cfg.getOption("Processes/%s/InFile" % process, None)

  def getProcesses(self):
    """Return the list of all processes available."""
    processesdict = self.cfg.getAsDict("Processes")
    processes = list(processesdict.keys())
    return processes

  def getProcessesDict(self):
    """Return all processes as a dictionary {'process':{'TarBall':Path, etc.

    etc.}}
    """
    return self.cfg.getAsDict("Processes")

  def existsProcess(self, process):
    """Check if the specified process exists."""
    return S_OK(self._existsProcess(process))

  def _existsProcess(self, process):
    """Check that the process exists."""
    return self.cfg.isSection('Processes/%s' % process)

  def writeProcessList(self, alternativePath=None):
    """Write the process list."""
    destination = self.location
    if alternativePath:
      destination = alternativePath
    written = self._writeProcessList(destination)
    if not written:
      return S_ERROR("Failed to write repository")
    return S_OK(destination)

  def printProcesses(self):
    """Dump to screen the content of the process list."""
    processesdict = self.cfg.getAsDict("Processes")
    pprint(processesdict)

  def uploadProcessListToFileCatalog(self, path_to_process_list, appVersion):
    """Upload the new processList to the FileCatalog."""
    from ILCDIRAC.Core.Utilities.FileUtils import upload
    from DIRAC.DataManagementSystem.Client.DataManager import DataManager
    from DIRAC import gConfig, exit as dexit

    datMan = DataManager()
    LOG.notice("Removing process list from file catalog" + path_to_process_list)
    res = datMan.removeFile(path_to_process_list)
    if not res['OK']:
      LOG.error("Could not remove process list from file catalog, do it by hand")
      dexit(2)
    LOG.notice("Done removing process list from file catalog")

    res = upload(os.path.dirname(path_to_process_list) + "/", self.location)
    if not res['OK']:
      LOG.error("something went wrong in the copy")
      dexit(2)

    LOG.notice("Putting process list to local processlist directory")
    localprocesslistpath = gConfig.getOption("/LocalSite/ProcessListPath", "")
    if localprocesslistpath['Value']:

      try:
        localSvnRepo = "/afs/cern.ch/eng/clic/software/whizard/whizard_195/"
        # because it does not make a difference if we hardcode it here or in
        # ${DIRAC}/etc/dirac.cfg, yours truly APS, JFS
        shutil.copy(self.location, localSvnRepo)
      except OSError as err:
        LOG.error("Copy of process list to %s failed with error %s!" % (localSvnRepo, str(err)))

      try:
        subprocess.call(["svn", "ci", os.path.join(localSvnRepo, os.path.basename(
            localprocesslistpath['Value'])), "-m'Process list for whizard version %s'" % appVersion], shell=False)
      except OSError as err:
        LOG.error("Commit failed! Error: %s" % str(err))

      try:
        shutil.copy(self.location, localprocesslistpath['Value'])
      except OSError as err:
        LOG.error("Copy of process list to %s failed!" % localprocesslistpath['Value'])

    LOG.notice("Done")
